from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.constants.relationship_label import RELATIONSHIP_TO_VALUE_LABEL
from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

    from .model import NodeUniquenessQueryRequest


class NodeUniqueAttributeConstraintQuery(Query):
    name = "node_constraints_uniqueness"
    insert_return = False
    type = QueryType.READ
    attribute_property_map = {"value": RELATIONSHIP_TO_VALUE_LABEL}

    def __init__(
        self,
        query_request: NodeUniquenessQueryRequest,
        min_count_required: int = 1,
        **kwargs: Any,
    ) -> None:
        self.query_request = query_request
        self.min_count_required = min_count_required
        super().__init__(**kwargs)

    def get_context(self) -> dict[str, str]:
        return {"kind": self.query_request.kind}

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string(), is_isolated=False)
        self.params.update(branch_params)
        from_times = db.render_list_comprehension(items="relationships(potential_path)", item_name="from")
        branch_name_and_level = db.render_list_comprehension_with_list(
            items="relationships(active_path)", item_names=["branch", "branch_level"]
        )

        attribute_names = set()
        attr_paths, attr_paths_with_value = [], []
        for attr_path in self.query_request.unique_attribute_paths:
            try:
                property_rel_name = self.attribute_property_map[attr_path.property_name or "value"]
            except KeyError as exc:
                raise ValueError(
                    f"{attr_path.property_name} is not a valid property for a uniqueness constraint"
                ) from exc
            attribute_names.add(attr_path.attribute_name)
            if attr_path.value:
                attr_paths_with_value.append((attr_path.attribute_name, property_rel_name, attr_path.value))
            else:
                attr_paths.append((attr_path.attribute_name, property_rel_name))

        relationship_names = set()
        relationship_attr_paths = []
        relationship_only_attr_paths = []
        relationship_attr_paths_with_value = []
        for rel_path in self.query_request.relationship_attribute_paths:
            relationship_names.add(rel_path.identifier)
            if rel_path.attribute_name and rel_path.value:
                relationship_attr_paths_with_value.append(
                    (rel_path.identifier, rel_path.attribute_name, rel_path.value)
                )
            elif rel_path.attribute_name:
                relationship_attr_paths.append((rel_path.identifier, rel_path.attribute_name))
            else:
                relationship_only_attr_paths.append(rel_path.identifier)

        if (
            not attr_paths
            and not attr_paths_with_value
            and not relationship_attr_paths
            and not relationship_attr_paths_with_value
            and not relationship_only_attr_paths
        ):
            raise ValueError(
                "The NodeUniquenessQueryRequest provided for node_constraints_uniqueness doesn't have enough information to continue"
            )

        self.params.update(
            {
                "node_kind": self.query_request.kind,
                "attr_paths": attr_paths,
                "attr_paths_with_value": attr_paths_with_value,
                "attribute_names": list(attribute_names),
                "relationship_names": list(relationship_names),
                "relationship_attr_paths": relationship_attr_paths,
                "relationship_attr_paths_with_value": relationship_attr_paths_with_value,
                "relationship_only_attr_paths": relationship_only_attr_paths,
                "min_count_required": self.min_count_required,
            }
        )

        attr_paths_subquery = """
        WITH start_node
        MATCH attr_path = (start_node)-[:HAS_ATTRIBUTE]->(attr:Attribute)-[r:HAS_VALUE]->(attr_value:AttributeValue)
        WHERE attr.name in $attribute_names
            AND ([attr.name, type(r)] in $attr_paths
            OR [attr.name, type(r), attr_value.value] in $attr_paths_with_value)
        RETURN attr_path as potential_path, NULL as rel_identifier, attr.name as potential_attr, attr_value.value as potential_attr_value
        """

        relationship_attr_paths_with_value_subquery = """
        WITH start_node
        MATCH rel_path = (start_node)-[:IS_RELATED]-(relationship_node:Relationship)-[:IS_RELATED]-(related_n:Node)-[:HAS_ATTRIBUTE]->(rel_attr:Attribute)-[:HAS_VALUE]->(rel_attr_value:AttributeValue)
        WHERE relationship_node.name in $relationship_names
            AND ([relationship_node.name, rel_attr.name] in $relationship_attr_paths
            OR [relationship_node.name, rel_attr.name, rel_attr_value.value] in $relationship_attr_paths_with_value)
        RETURN rel_path as potential_path, relationship_node.name as rel_identifier, rel_attr.name as potential_attr, rel_attr_value.value as potential_attr_value
        """

        relationship_only_attr_paths_subquery = """
        WITH start_node
        MATCH rel_path = (start_node)-[:IS_RELATED]-(relationship_node:Relationship)-[:IS_RELATED]-(related_n:Node)
        WHERE relationship_node.name in $relationship_only_attr_paths
        RETURN rel_path as potential_path, relationship_node.name as rel_identifier, "id" as potential_attr, related_n.uuid as potential_attr_value
        """

        select_subqueries = []
        if attr_paths or attr_paths_with_value:
            select_subqueries.append(attr_paths_subquery)
        if relationship_attr_paths_with_value or relationship_attr_paths:
            select_subqueries.append(relationship_attr_paths_with_value_subquery)
        if relationship_only_attr_paths:
            select_subqueries.append(relationship_only_attr_paths_subquery)

        select_subqueries_str = "UNION".join(select_subqueries)

        # ruff: noqa: E501
        query = """
        // group by node
        MATCH (start_node:%(node_kind)s)
        // get attributes for node and its relationships
        CALL {
            %(select_subqueries_str)s
        }
        CALL {
            WITH potential_path
            WITH potential_path  // workaround for neo4j not allowing WHERE in a WITH of a subquery
            // only the branches and times we care about
            WHERE all(
                r IN relationships(potential_path) WHERE (
                    %(branch_filter)s
                )
            )
            // only get the latest path on the farthest branch from main
            RETURN
                potential_path as matched_path,
                reduce(br_lvl = 0, r in relationships(potential_path) | br_lvl + r.branch_level) AS branch_level_sum,
                %(from_times)s AS from_times,
                // used as tiebreaker for updated relationships that were deleted and added at the same microsecond
                reduce(active_count = 0, r in relationships(potential_path) | active_count + (CASE r.status WHEN "active" THEN 1 ELSE 0 END)) AS active_relationship_count
        }
        WITH
            collect([matched_path, branch_level_sum, from_times, active_relationship_count, potential_attr_value]) as enriched_paths,
            start_node,
            rel_identifier,
            potential_attr
        CALL {
            WITH enriched_paths
            UNWIND enriched_paths as path_to_check
            RETURN path_to_check[0] as current_path, path_to_check[4] as latest_value
            ORDER BY
                path_to_check[1] DESC,
                path_to_check[2][-1] DESC,
                path_to_check[2][-2] DESC,
                path_to_check[3] DESC
            LIMIT 1
        }
        CALL {
            // only active paths
            WITH current_path
            WITH current_path  // workaround for neo4j not allowing WHERE in a WITH of a subquery
            WHERE all(r IN relationships(current_path) WHERE r.status = "active")
            RETURN current_path as active_path
        }
        CALL {
            // get deepest branch name
            WITH active_path
            UNWIND %(branch_name_and_level)s as branch_name_and_level
            RETURN branch_name_and_level[0] as branch_name
            ORDER BY branch_name_and_level[1] DESC
            LIMIT 1
        }
        // only duplicate values
        WITH
            collect([start_node.uuid, branch_name]) as nodes_and_branches,
            count(*) as node_count,
            potential_attr as attr_name,
            latest_value as attr_value,
            rel_identifier as relationship_identifier
        WHERE node_count > $min_count_required
        UNWIND nodes_and_branches as node_and_branch
        RETURN
            node_and_branch[0] as node_id,
            node_and_branch[1] as deepest_branch_name,
            node_count,
            attr_name,
            attr_value,
            relationship_identifier
        """ % {
            "node_kind": self.query_request.kind,
            "select_subqueries_str": select_subqueries_str,
            "branch_filter": branch_filter,
            "from_times": from_times,
            "branch_name_and_level": branch_name_and_level,
        }

        self.add_to_query(query)
        self.return_labels = [
            "node_id",
            "deepest_branch_name",
            "node_count",
            "attr_name",
            "attr_value",
            "relationship_identifier",
        ]



from pydantic import ConfigDict, Field

from infrahub.generators.models import ProposedChangeGeneratorDefinition
from infrahub.message_bus import InfrahubMessage
from infrahub.message_bus.types import ProposedChangeBranchDiff


class RequestGeneratorDefinitionCheck(InfrahubMessage):
    """Sent to trigger Generators to run for a proposed change."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    generator_definition: ProposedChangeGeneratorDefinition = Field(..., description="The Generator Definition")
    branch_diff: ProposedChangeBranchDiff = Field(..., description="The calculated diff between the two branches")
    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    source_branch: str = Field(..., description="The source branch")
    source_branch_sync_with_git: bool = Field(..., description="Indicates if the source branch should sync with git")
    destination_branch: str = Field(..., description="The target branch")



from .builder.constraint.grouped.node_runner import NodeConstraintRunnerDependency
from .builder.constraint.node.grouped_uniqueness import NodeGroupedUniquenessConstraintDependency
from .builder.constraint.node.uniqueness import NodeAttributeUniquenessConstraintDependency
from .builder.constraint.relationship_manager.count import RelationshipCountConstraintDependency
from .builder.constraint.relationship_manager.peer_kind import RelationshipPeerKindConstraintDependency
from .builder.constraint.relationship_manager.profiles_kind import RelationshipProfilesKindConstraintDependency
from .builder.constraint.schema.aggregated import AggregatedSchemaConstraintsDependency
from .builder.constraint.schema.attribute_regex import SchemaAttributeRegexConstraintDependency
from .builder.constraint.schema.attribute_uniqueness import SchemaAttributeUniqueConstraintDependency
from .builder.constraint.schema.relationship_optional import SchemaRelationshipOptionalConstraintDependency
from .builder.constraint.schema.uniqueness import SchemaUniquenessConstraintDependency
from .builder.diff.calculator import DiffCalculatorDependency
from .builder.diff.combiner import DiffCombinerDependency
from .builder.diff.conflict_transferer import DiffConflictTransfererDependency
from .builder.diff.coordinator import DiffCoordinatorDependency
from .builder.diff.data_check_synchronizer import DiffDataCheckSynchronizerDependency
from .builder.diff.diff_merger import DiffMergerDependency
from .builder.diff.enricher.aggregated import DiffAggregatedEnricherDependency
from .builder.diff.enricher.cardinality_one import DiffCardinalityOneEnricherDependency
from .builder.diff.enricher.hierarchy import DiffHierarchyEnricherDependency
from .builder.diff.ipam_diff_parser import IpamDiffParserDependency
from .builder.diff.repository import DiffRepositoryDependency
from .builder.ip.kinds_getter import IpamKindsGetterDependency
from .builder.node.delete_validator import NodeDeleteValidatorDependency
from .component.registry import ComponentDependencyRegistry


def build_component_registry() -> ComponentDependencyRegistry:
    component_registry = ComponentDependencyRegistry.get_registry()
    component_registry.track_dependency(AggregatedSchemaConstraintsDependency)
    component_registry.track_dependency(SchemaAttributeRegexConstraintDependency)
    component_registry.track_dependency(SchemaAttributeUniqueConstraintDependency)
    component_registry.track_dependency(SchemaRelationshipOptionalConstraintDependency)
    component_registry.track_dependency(SchemaUniquenessConstraintDependency)
    component_registry.track_dependency(NodeAttributeUniquenessConstraintDependency)
    component_registry.track_dependency(NodeGroupedUniquenessConstraintDependency)
    component_registry.track_dependency(RelationshipCountConstraintDependency)
    component_registry.track_dependency(RelationshipProfilesKindConstraintDependency)
    component_registry.track_dependency(RelationshipPeerKindConstraintDependency)
    component_registry.track_dependency(NodeConstraintRunnerDependency)
    component_registry.track_dependency(NodeDeleteValidatorDependency)
    component_registry.track_dependency(IpamKindsGetterDependency)
    component_registry.track_dependency(IpamDiffParserDependency)
    component_registry.track_dependency(DiffCardinalityOneEnricherDependency)
    component_registry.track_dependency(DiffHierarchyEnricherDependency)
    component_registry.track_dependency(DiffAggregatedEnricherDependency)
    component_registry.track_dependency(DiffCalculatorDependency)
    component_registry.track_dependency(DiffCombinerDependency)
    component_registry.track_dependency(DiffRepositoryDependency)
    component_registry.track_dependency(DiffConflictTransfererDependency)
    component_registry.track_dependency(DiffCoordinatorDependency)
    component_registry.track_dependency(DiffDataCheckSynchronizerDependency)
    component_registry.track_dependency(DiffMergerDependency)
    return component_registry


def get_component_registry() -> ComponentDependencyRegistry:
    return ComponentDependencyRegistry.get_registry()



from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class RequestGeneratorRun(BaseModel):
    """Runs a generator."""

    generator_definition: ProposedChangeGeneratorDefinition = Field(..., description="The Generator definition")
    generator_instance: Optional[str] = Field(
        default=None, description="The id of the generator instance if it previously existed"
    )
    commit: str = Field(..., description="The commit to target")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    repository_kind: str = Field(..., description="The kind of the Repository")
    branch_name: str = Field(..., description="The branch where the check is run")
    target_id: str = Field(..., description="The ID of the target object for this generator")
    target_name: str = Field(..., description="Name of the generator target")
    query: str = Field(..., description="The name of the query to use when collecting data")
    variables: dict = Field(..., description="Input variables when running the generator")


class RequestGeneratorDefinitionRun(BaseModel):
    """Sent to trigger a Generator to run on a specific branch."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    generator_definition: ProposedChangeGeneratorDefinition = Field(..., description="The Generator Definition")
    branch: str = Field(..., description="The branch to target")


class ProposedChangeGeneratorDefinition(BaseModel):
    definition_id: str
    definition_name: str
    query_name: str
    convert_query_response: bool
    query_models: list[str]
    repository_id: str
    class_name: str
    file_path: str
    parameters: dict
    group_id: str



from typing import Any

from neo4j.graph import Node as Neo4jNode

from infrahub.core.diff.model.path import ConflictSelection
from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase


class EnrichedDiffConflictUpdateQuery(Query):
    name = "enriched_diff_conflict_update"
    type = QueryType.WRITE

    def __init__(
        self,
        conflict_id: str,
        selection: ConflictSelection | None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.conflict_id = conflict_id
        self.selection = selection

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        self.params = {"conflict_id": self.conflict_id, "selection": self.selection.value if self.selection else None}
        query = """
        MATCH (conflict:DiffConflict {uuid: $conflict_id})
        SET conflict.selected_branch = $selection
        """
        self.return_labels = ["conflict"]
        self.add_to_query(query=query)

    async def get_conflict_node(self) -> Neo4jNode | None:
        result = self.get_result()
        if not result:
            return None
        return result.get_node("conflict")



from infrahub_sdk.exceptions import ModuleImportError
from infrahub_sdk.node import InfrahubNode
from infrahub_sdk.schema.repository import InfrahubGeneratorDefinitionConfig
from prefect import flow
from prefect.logging import get_run_logger

from infrahub import lock
from infrahub.core.constants import GeneratorInstanceStatus, InfrahubKind, ValidatorConclusion
from infrahub.core.timestamp import Timestamp
from infrahub.git.base import extract_repo_file_information
from infrahub.git.repository import get_initialized_repo
from infrahub.message_bus import messages
from infrahub.services import InfrahubServices
from infrahub.tasks.check import set_check_status
from infrahub.workflows.utils import add_tags

# pylint: disable=duplicate-code


@flow(
    name="git-repository-check-generator-run",
    flow_run_name="Execute Generator {message.generator_definition.definition_name} for {message.target_name}",
)
async def run(message: messages.CheckGeneratorRun, service: InfrahubServices) -> None:
    if message.proposed_change:
        await add_tags(branches=[message.branch_name], nodes=[message.proposed_change], db_change=True)
    else:
        await add_tags(branches=[message.branch_name], nodes=[message.repository_id], db_change=True)

    log = get_run_logger()

    repository = await get_initialized_repo(
        repository_id=message.repository_id,
        name=message.repository_name,
        service=service,
        repository_kind=message.repository_kind,
        commit=message.commit,
    )

    conclusion = ValidatorConclusion.SUCCESS

    generator_definition = InfrahubGeneratorDefinitionConfig(
        name=message.generator_definition.definition_name,
        class_name=message.generator_definition.class_name,
        file_path=message.generator_definition.file_path,
        query=message.generator_definition.query_name,
        targets=message.generator_definition.group_id,
        convert_query_response=message.generator_definition.convert_query_response,
    )

    commit_worktree = repository.get_commit_worktree(commit=message.commit)

    file_info = extract_repo_file_information(
        full_filename=commit_worktree.directory / generator_definition.file_path,
        repo_directory=repository.directory_root,
        worktree_directory=commit_worktree.directory,
    )
    generator_instance = await _define_instance(message=message, service=service)

    check_message = "Instance successfully generated"
    try:
        log.debug(f"repo information {file_info}")
        log.debug(f"Root directory : {repository.directory_root}")
        generator_class = generator_definition.load_class(
            import_root=repository.directory_root, relative_path=file_info.relative_repo_path_dir
        )

        generator = generator_class(
            query=generator_definition.query,
            client=service.client,
            branch=message.branch_name,
            params=message.variables,
            generator_instance=generator_instance.id,
            convert_query_response=generator_definition.convert_query_response,
            infrahub_node=InfrahubNode,
        )
        await generator.run(identifier=generator_definition.name)
        generator_instance.status.value = GeneratorInstanceStatus.READY.value
    except ModuleImportError as exc:
        conclusion = ValidatorConclusion.FAILURE
        generator_instance.status.value = GeneratorInstanceStatus.ERROR.value
        check_message = f"Failed to import generator: {exc.message}"
        log.exception(check_message, exc_info=exc)
    except Exception as exc:  # pylint: disable=broad-exception-caught
        conclusion = ValidatorConclusion.FAILURE
        generator_instance.status.value = GeneratorInstanceStatus.ERROR.value
        check_message = f"Failed to execute generator: {str(exc)}"
        log.exception(check_message, exc_info=exc)

    log.info("Generator run completed, starting update")
    await generator_instance.update(do_full_update=True)

    check = None
    existing_check = await service.client.filters(
        kind=InfrahubKind.GENERATORCHECK, validator__ids=message.validator_id, instance__value=generator_instance.id
    )
    if existing_check:
        check = existing_check[0]

    if check:
        check.created_at.value = Timestamp().to_string()
        check.conclusion.value = conclusion.value
        await check.save()
    else:
        check = await service.client.create(
            kind=InfrahubKind.GENERATORCHECK,
            data={
                "name": message.target_name,
                "origin": message.repository_id,
                "kind": "GeneratorDefinition",
                "validator": message.validator_id,
                "created_at": Timestamp().to_string(),
                "message": check_message,
                "conclusion": conclusion.value,
                "instance": generator_instance.id,
            },
        )
        await check.save()

    await set_check_status(message=message, conclusion=conclusion.value, service=service)


async def _define_instance(message: messages.CheckGeneratorRun, service: InfrahubServices) -> InfrahubNode:
    if message.generator_instance:
        instance = await service.client.get(
            kind=InfrahubKind.GENERATORINSTANCE, id=message.generator_instance, branch=message.branch_name
        )
        instance.status.value = GeneratorInstanceStatus.PENDING.value
        await instance.update(do_full_update=True)

    else:
        async with lock.registry.get(
            f"{message.target_id}-{message.generator_definition.definition_id}", namespace="generator"
        ):
            instances = await service.client.filters(
                kind=InfrahubKind.GENERATORINSTANCE,
                definition__ids=[message.generator_definition.definition_id],
                object__ids=[message.target_id],
                branch=message.branch_name,
            )
            if instances:
                instance = instances[0]
                instance.status.value = GeneratorInstanceStatus.PENDING.value
                await instance.update(do_full_update=True)
            else:
                instance = await service.client.create(
                    kind=InfrahubKind.GENERATORINSTANCE,
                    branch=message.branch_name,
                    data={
                        "name": f"{message.generator_definition.definition_name}: {message.target_name}",
                        "status": GeneratorInstanceStatus.PENDING.value,
                        "object": message.target_id,
                        "definition": message.generator_definition.definition_id,
                    },
                )
                await instance.save()
    return instance



from pydantic import ConfigDict, Field

from infrahub.message_bus import InfrahubMessage
from infrahub.message_bus.types import ProposedChangeBranchDiff


class CheckRepositoryUserCheck(InfrahubMessage):
    """Runs a check as defined within a CoreCheckDefinition within a repository."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    validator_id: str = Field(..., description="The id of the validator associated with this check")
    validator_execution_id: str = Field(..., description="The id of current execution of the associated validator")
    check_execution_id: str = Field(..., description="The unique ID for the current execution of this check")
    check_definition_id: str = Field(..., description="The unique ID of the check definition")
    commit: str = Field(..., description="The commit to target")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    branch_name: str = Field(..., description="The branch where the check is run")
    file_path: str = Field(..., description="The path and filename of the check")
    class_name: str = Field(..., description="The name of the class containing the check")
    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    variables: dict = Field(default_factory=dict, description="Input variables when running the check")
    name: str = Field(..., description="The name of the check")
    branch_diff: ProposedChangeBranchDiff = Field(..., description="The calculated diff between the two branches")



from typing import Any

from pydantic import Field

from infrahub.message_bus import InfrahubMessage


class EventNodeMutated(InfrahubMessage):
    """Sent when a node has been mutated"""

    branch: str = Field(..., description="The branch that was created")
    kind: str = Field(..., description="The type of object modified")
    node_id: str = Field(..., description="The ID of the mutated node")
    action: str = Field(..., description="The action taken on the node")
    data: dict[str, Any] = Field(..., description="Data on modified object")



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core import registry
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.core.migrations.shared import MigrationResult
from infrahub.dependencies.registry import build_component_registry, get_component_registry
from infrahub.log import get_logger

from ..shared import ArbitraryMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class Migration015(ArbitraryMigration):
    name: str = "015_diff_format_update"
    minimum_version: int = 14

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        result = MigrationResult()

        return result

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        default_branch = registry.get_branch_from_registry()
        build_component_registry()
        component_registry = get_component_registry()
        diff_repo = await component_registry.get_component(DiffRepository, db=db, branch=default_branch)

        diff_roots = await diff_repo.get_roots_metadata()
        await diff_repo.delete_diff_roots(diff_root_uuids=[d.uuid for d in diff_roots])
        return MigrationResult()



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

import jinja2

from . import protocols as sdk_protocols
from .ctl.constants import PROTOCOLS_TEMPLATE
from .schema import (
    AttributeSchemaAPI,
    GenericSchema,
    GenericSchemaAPI,
    MainSchemaTypesAll,
    NodeSchema,
    NodeSchemaAPI,
    ProfileSchemaAPI,
    RelationshipSchemaAPI,
)

ATTRIBUTE_KIND_MAP = {
    "ID": "String",
    "Text": "String",
    "TextArea": "String",
    "DateTime": "DateTime",
    "Email": "String",
    "Password": "String",
    "HashedPassword": "HashedPassword",
    "URL": "URL",
    "File": "String",
    "MacAddress": "MacAddress",
    "Color": "String",
    "Dropdown": "Dropdown",
    "Number": "Integer",
    "Bandwidth": "Integer",
    "IPHost": "IPHost",
    "IPNetwork": "IPNetwork",
    "Boolean": "Boolean",
    "Checkbox": "Boolean",
    "List": "ListAttribute",
    "JSON": "JSONAttribute",
    "Any": "AnyAttribute",
}


class CodeGenerator:
    def __init__(self, schema: dict[str, MainSchemaTypesAll]):
        self.generics: dict[str, GenericSchemaAPI | GenericSchema] = {}
        self.nodes: dict[str, NodeSchemaAPI | NodeSchema] = {}
        self.profiles: dict[str, ProfileSchemaAPI] = {}

        for name, schema_type in schema.items():
            if isinstance(schema_type, (GenericSchemaAPI, GenericSchema)):
                self.generics[name] = schema_type
            if isinstance(schema_type, (NodeSchemaAPI, NodeSchema)):
                self.nodes[name] = schema_type
            if isinstance(schema_type, ProfileSchemaAPI):
                self.profiles[name] = schema_type

        self.base_protocols = [
            e
            for e in dir(sdk_protocols)
            if not e.startswith("__")
            and not e.endswith("__")
            and e
            not in ("TYPE_CHECKING", "CoreNode", "Optional", "Protocol", "Union", "annotations", "runtime_checkable")
        ]

        self.sorted_generics = self._sort_and_filter_models(self.generics, filters=["CoreNode"] + self.base_protocols)
        self.sorted_nodes = self._sort_and_filter_models(self.nodes, filters=["CoreNode"] + self.base_protocols)
        self.sorted_profiles = self._sort_and_filter_models(
            self.profiles, filters=["CoreProfile"] + self.base_protocols
        )

    def render(self, sync: bool = True) -> str:
        jinja2_env = jinja2.Environment(loader=jinja2.BaseLoader(), trim_blocks=True, lstrip_blocks=True)
        jinja2_env.filters["inheritance"] = self._jinja2_filter_inheritance
        jinja2_env.filters["render_attribute"] = self._jinja2_filter_render_attribute
        jinja2_env.filters["render_relationship"] = self._jinja2_filter_render_relationship

        template = jinja2_env.from_string(PROTOCOLS_TEMPLATE)
        return template.render(
            generics=self.sorted_generics,
            nodes=self.sorted_nodes,
            profiles=self.sorted_profiles,
            base_protocols=self.base_protocols,
            sync=sync,
        )

    @staticmethod
    def _jinja2_filter_inheritance(value: dict[str, Any]) -> str:
        inherit_from: list[str] = value.get("inherit_from", [])

        if not inherit_from:
            return "CoreNode"
        return ", ".join(inherit_from)

    @staticmethod
    def _jinja2_filter_render_attribute(value: AttributeSchemaAPI) -> str:
        attribute_kind: str = ATTRIBUTE_KIND_MAP[value.kind]

        if value.optional:
            attribute_kind += "Optional"

        return f"{value.name}: {attribute_kind}"

    @staticmethod
    def _jinja2_filter_render_relationship(value: RelationshipSchemaAPI, sync: bool = False) -> str:
        name = value.name
        cardinality = value.cardinality

        type_ = "RelatedNode"
        if cardinality == "many":
            type_ = "RelationshipManager"

        if sync:
            type_ += "Sync"

        return f"{name}: {type_}"

    @staticmethod
    def _sort_and_filter_models(
        models: Mapping[str, MainSchemaTypesAll], filters: list[str] | None = None
    ) -> list[MainSchemaTypesAll]:
        if filters is None:
            filters = ["CoreNode"]

        filtered: list[MainSchemaTypesAll] = []
        for name, model in models.items():
            if name in filters:
                continue
            filtered.append(model)

        return sorted(filtered, key=lambda k: k.name)



from typing import Any, Optional

from pydantic import ConfigDict, Field

from infrahub.core.constants import TaskConclusion
from infrahub.core.node.standard import StandardNode
from infrahub.core.protocols import CoreNode
from infrahub.core.query.standard_node import StandardNodeQuery
from infrahub.core.query.task import TaskNodeCreateQuery, TaskNodeQuery, TaskNodeQueryWithLogs
from infrahub.core.timestamp import current_timestamp
from infrahub.database import InfrahubDatabase
from infrahub.utils import get_nested_dict

from .task_log import TaskLog


class Task(StandardNode):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    title: str
    conclusion: TaskConclusion
    account_id: Optional[str] = Field(default=None, description="The ID of the account that created this task")
    created_at: str = Field(default_factory=current_timestamp, description="The time when this task was created")
    updated_at: str = Field(default_factory=current_timestamp, description="The time when this task was last updated")
    related_node: Optional[CoreNode] = Field(default=None, description="The Infrahub node that this object refers to")

    _exclude_attrs: list[str] = ["id", "uuid", "account_id", "_query", "related_node"]
    _query: type[StandardNodeQuery] = TaskNodeCreateQuery

    @property
    def related(self) -> CoreNode:
        if self.related_node:
            return self.related_node
        raise ValueError("The related_node field has not been populated")

    @classmethod
    async def query(
        cls,
        db: InfrahubDatabase,
        fields: dict[str, Any],
        limit: int | None = None,
        offset: int | None = None,
        ids: list[str] | None = None,
        related_nodes: list[str] | None = None,
    ) -> dict[str, Any]:
        ids = ids or []
        related_nodes = related_nodes or []
        log_fields = get_nested_dict(nested_dict=fields, keys=["edges", "node", "logs", "edges", "node"])
        count = None
        if "count" in fields:
            query = await TaskNodeQuery.init(db=db, ids=ids, related_nodes=related_nodes)
            count = await query.count(db=db)

        if log_fields:
            query = await TaskNodeQueryWithLogs.init(
                db=db, limit=limit, offset=offset, ids=ids, related_nodes=related_nodes
            )
            await query.execute(db=db)
        else:
            query = await TaskNodeQuery.init(db=db, limit=limit, offset=offset, ids=ids, related_nodes=related_nodes)
            await query.execute(db=db)

        nodes: list[dict] = []
        for result in query.get_results():
            related_node = result.get("rn")
            task_result = result.get_node("n")
            logs = []
            if log_fields:
                logs_results = result.get_node_collection("logs")
                logs = [
                    {
                        "node": await TaskLog.from_db(result, extras={"task_id": task_result.get("uuid")}).to_graphql(
                            fields=log_fields
                        )
                    }
                    for result in logs_results
                ]

            task = cls.from_db(task_result)
            nodes.append(
                {
                    "node": {
                        "title": task.title,
                        "conclusion": task.conclusion,
                        "related_node": related_node.get("uuid"),
                        "related_node_kind": related_node.get("kind"),
                        "created_at": task.created_at,
                        "updated_at": task.updated_at,
                        "id": task_result.get("uuid"),
                        "logs": {"edges": logs},
                    }
                }
            )

        return {"count": count, "edges": nodes}



from prefect import flow

from infrahub.log import get_logger
from infrahub.message_bus import messages
from infrahub.services import InfrahubServices

log = get_logger()


@flow(name="event-schema-update")
async def update(message: messages.EventSchemaUpdate, service: InfrahubServices) -> None:
    log.info("run_message", branch=message.branch)

    msg = messages.RefreshRegistryBranches()

    msg.assign_meta(parent=message)
    await service.send(message=msg)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from infrahub.core.constants import GLOBAL_BRANCH_NAME, BranchSupportType
from infrahub.core.query import Query, QueryResult, QueryType, sort_results_by_time
from infrahub.core.timestamp import Timestamp

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase


class DiffQuery(Query):
    branch_names: list[str]
    diff_from: Timestamp
    diff_to: Timestamp
    type: QueryType = QueryType.READ

    def __init__(
        self,
        branch: Branch,
        diff_from: Union[Timestamp, str] = None,
        diff_to: Union[Timestamp, str] = None,
        **kwargs,
    ):
        """A diff is always in the context of a branch"""

        if not diff_from and branch.is_default:
            raise ValueError("diff_from is mandatory when the diff is on the main branch.")

        # If diff from hasn't been provided, we'll use the creation of the branch as the starting point
        if diff_from:
            self.diff_from = Timestamp(diff_from)
        else:
            self.diff_from = Timestamp(branch.created_at)

        # If Diff_to is not defined it will automatically select the current time.
        self.diff_to = Timestamp(diff_to)

        if self.diff_to < self.diff_from:
            raise ValueError("diff_to must be later than diff_from")

        self.branch_names = branch.get_branches_in_scope()

        super().__init__(branch, **kwargs)


class DiffNodeQuery(DiffQuery):
    name = "diff_node"
    type = QueryType.READ

    def __init__(
        self,
        namespaces_include: Optional[list[str]] = None,
        namespaces_exclude: Optional[list[str]] = None,
        kinds_include: Optional[list[str]] = None,
        kinds_exclude: Optional[list[str]] = None,
        branch_support: Optional[list[BranchSupportType]] = None,
        **kwargs,
    ):
        self.namespaces_include = namespaces_include
        self.namespaces_exclude = namespaces_exclude
        self.kinds_include = kinds_include
        self.kinds_exclude = kinds_exclude
        self.branch_support = branch_support or [BranchSupportType.AWARE]

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        # TODO need to improve the query to capture an object that has been deleted into the branch
        # TODO probably also need to consider a node what was merged already

        br_filter, br_params = self.branch.get_query_filter_range(
            rel_label="r",
            start_time=self.diff_from,
            end_time=self.diff_to,
        )

        self.params.update(br_params)
        self.params["branch_support"] = [item.value for item in self.branch_support]

        where_clause = ""
        if self.namespaces_include:
            where_clause += "n.namespace IN $namespaces_include AND "
            self.params["namespaces_include"] = self.namespaces_include

        if self.namespaces_exclude:
            where_clause += "NOT(n.namespace IN $namespaces_exclude) AND "
            self.params["namespaces_exclude"] = self.namespaces_exclude

        if self.kinds_include:
            where_clause += "n.kind IN $kinds_include AND "
            self.params["kinds_include"] = self.kinds_include

        if self.kinds_exclude:
            where_clause += "NOT(n.kind IN $kinds_exclude) AND "
            self.params["kinds_exclude"] = self.kinds_exclude

        where_clause += "n.branch_support IN $branch_support AND %s" % "\n AND ".join(br_filter)

        query = (
            """
        MATCH (root:Root)-[r:IS_PART_OF]-(n:Node)
        WHERE %s
        """
            % where_clause
        )

        self.add_to_query(query)
        self.order_by = ["n.uuid"]

        self.return_labels = ["n", "r"]


class DiffAttributeQuery(DiffQuery):
    name = "diff_attribute"
    type = QueryType.READ

    def __init__(
        self,
        namespaces_include: Optional[list[str]] = None,
        namespaces_exclude: Optional[list[str]] = None,
        kinds_include: Optional[list[str]] = None,
        kinds_exclude: Optional[list[str]] = None,
        branch_support: Optional[list[BranchSupportType]] = None,
        **kwargs,
    ):
        self.namespaces_include = namespaces_include
        self.namespaces_exclude = namespaces_exclude
        self.kinds_include = kinds_include
        self.kinds_exclude = kinds_exclude
        self.branch_support = branch_support or [BranchSupportType.AWARE]

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        # TODO need to improve the query to capture an object that has been deleted into the branch

        rels_filters, rels_params = self.branch.get_query_filter_relationships_diff(
            rel_labels=["r2"], diff_from=self.diff_from, diff_to=self.diff_to
        )

        self.params.update(rels_params)
        self.params["branch_support"] = [item.value for item in self.branch_support]

        where_clause = ""
        if self.namespaces_include:
            where_clause += "n.namespace IN $namespaces_include AND "
            self.params["namespaces_include"] = self.namespaces_include

        if self.namespaces_exclude:
            where_clause += "NOT(n.namespace IN $namespaces_exclude) AND "
            self.params["namespaces_exclude"] = self.namespaces_exclude

        if self.kinds_include:
            where_clause += "n.kind IN $kinds_include AND "
            self.params["kinds_include"] = self.kinds_include

        if self.kinds_exclude:
            where_clause += "NOT(n.kind IN $kinds_exclude) AND "
            self.params["kinds_exclude"] = self.kinds_exclude

        where_clause += "a.branch_support IN $branch_support AND %s" % "\n AND ".join(rels_filters)

        query = (
            """
        MATCH (n:Node)-[r1:HAS_ATTRIBUTE]-(a:Attribute)-[r2:HAS_VALUE|IS_VISIBLE|IS_PROTECTED|HAS_SOURCE|HAS_OWNER]->(ap)
        WHERE %s
        """
            % where_clause
        )

        self.add_to_query(query)

        self.return_labels = ["n", "a", "ap", "r1", "r2"]


class DiffRelationshipQuery(DiffQuery):
    name = "diff_relationship"
    type = QueryType.READ

    def __init__(
        self,
        namespaces_include: Optional[list[str]] = None,
        namespaces_exclude: Optional[list[str]] = None,
        kinds_include: Optional[list[str]] = None,
        kinds_exclude: Optional[list[str]] = None,
        branch_support: Optional[list[BranchSupportType]] = None,
        **kwargs,
    ):
        self.namespaces_include = namespaces_include
        self.namespaces_exclude = namespaces_exclude
        self.kinds_include = kinds_include
        self.kinds_exclude = kinds_exclude
        self.branch_support = branch_support or [BranchSupportType.AWARE]

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        where_clause = ""
        if self.namespaces_include:
            where_clause += "(src.namespace IN $namespaces_include OR dst.namespace IN $namespaces_include) AND "
            self.params["namespaces_include"] = self.namespaces_include

        if self.namespaces_exclude:
            where_clause += "NOT(src.namespace IN $namespaces_exclude OR dst.namespace IN $namespaces_exclude) AND "
            self.params["namespaces_exclude"] = self.namespaces_exclude

        if self.kinds_include:
            where_clause += "(src.kind IN $kinds_include OR dst.kind IN $kinds_include) AND "
            self.params["kinds_include"] = self.kinds_include

        if self.kinds_exclude:
            where_clause += "NOT(src.kind IN $kinds_exclude OR dst.kind IN $kinds_exclude) AND "
            self.params["kinds_exclude"] = self.kinds_exclude

        query = (
            """
        CALL {
            MATCH p = (src:Node)-[r1:IS_RELATED]-(rel:Relationship)-[r2:IS_RELATED]-(dst:Node)
            WHERE (rel.branch_support IN $branch_support AND %s r1.branch = r2.branch AND
                (r1.to = r2.to OR (r1.to is NULL AND r2.to is NULL))
                AND r1.from = r2.from AND r1.status = r2.status
                AND r1.branch IN $branch_names
                AND (
                    (r1.from >= $diff_from AND r1.from <= $diff_to AND r1.to is NULL)
                    OR (r1.to >= $diff_from AND r1.to <= $diff_to)
                )
            )
            RETURN DISTINCT [rel.uuid, r1.branch] as identifier, rel, r1.branch as branch_name
        }
        CALL {
            WITH rel, branch_name
            MATCH p = (sn:Node)-[r1:IS_RELATED]-(rel:Relationship)-[r2:IS_RELATED]-(dn:Node)
            WHERE (rel.branch_support IN $branch_support AND r1.branch = r2.branch AND
                (r1.to = r2.to OR (r1.to is NULL AND r2.to is NULL))
                AND r1.from = r2.from AND r1.status = r2.status
                AND r1.branch = branch_name
                AND (
                    (r1.from >= $diff_from AND r1.from <= $diff_to AND r1.to is NULL)
                    OR (r1.to >= $diff_from AND r1.to <= $diff_to)
                )
                AND sn <> dn
            )
            RETURN rel as rel1, sn as sn1, dn as dn1, r1 as r11, r2 as r21
            ORDER BY r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH rel1 as rel, sn1 as sn, dn1 as dn, r11 as r1, r21 as r2
        """
            % where_clause
        )

        self.add_to_query(query)
        self.params["branch_names"] = self.branch_names
        self.params["diff_from"] = self.diff_from.to_string()
        self.params["diff_to"] = self.diff_to.to_string()
        self.params["branch_support"] = [item.value for item in self.branch_support]

        self.return_labels = ["sn", "dn", "rel", "r1", "r2"]


class DiffRelationshipPropertyQuery(DiffQuery):
    name = "diff_relationship_property"
    type = QueryType.READ

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        rels_filter, rels_params = self.branch.get_query_filter_relationships_range(
            rel_labels=["r"], start_time=self.diff_from, end_time=self.diff_to
        )
        self.params.update(rels_params)

        query = """
        CALL {
            MATCH (rel:Relationship)-[r3:IS_VISIBLE|IS_PROTECTED|HAS_SOURCE|HAS_OWNER]-()
            WHERE (
                r3.branch IN $branch_names
                AND (r3.from >= $diff_from AND r3.from <= $diff_to AND r3.to is NULL)
                OR (r3.to >= $diff_from AND r3.to <= $diff_to)
            )
            RETURN DISTINCT rel
        }
        CALL {
            WITH rel
            MATCH p = ((sn:Node)-[r1]-(rel)-[r2]-(dn:Node))
            WHERE r1.branch = r2.branch AND (r1.to = r2.to OR (r1.to is NULL AND r2.to is NULL))
            AND r1.from = r2.from AND r1.status = r2.status AND all(r IN relationships(p) WHERE ( %s ))
            AND sn <> dn
            RETURN rel as rel1, sn as sn1, dn as dn1, r1 as r11, r2 as r21
            ORDER BY r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH rel1 as rel, sn1 as sn, dn1 as dn, r11 as r1, r21 as r2
        CALL {
            // -----------------------
            // group results to the latest entry for each edge type (IS_VISIBLE, etc.)
            // -----------------------
            WITH rel
            MATCH (rel:Relationship)-[r3:IS_VISIBLE|IS_PROTECTED|HAS_SOURCE|HAS_OWNER]->(prop)
            WHERE (
                r3.branch IN $branch_names
                AND (r3.from >= $diff_from AND r3.from <= $diff_to AND r3.to is NULL)
                OR (r3.to >= $diff_from AND r3.to <= $diff_to)
            )
            WITH r3, prop
            ORDER BY r3.branch, type(r3), r3.from DESC
            WITH r3.branch AS r3_branch, type(r3) AS type_r3, head(collect([r3, prop])) AS data
            RETURN data[0] AS r3, data[1] AS rp
        }
        """ % "\n AND ".join(rels_filter)

        self.add_to_query(query)
        self.params["branch_names"] = self.branch_names
        self.params["diff_from"] = self.diff_from.to_string()
        self.params["diff_to"] = self.diff_to.to_string()

        self.return_labels = ["sn", "dn", "rel", "rp", "r3", "r1", "r2"]


class DiffNodePropertiesByIDSRangeQuery(Query):
    name = "diff_node_properties_range_ids"
    type = QueryType.READ

    def __init__(self, ids: list[str], diff_from: str, diff_to: str, account=None, **kwargs):
        self.account = account
        self.ids = ids
        self.time_from = Timestamp(diff_from)
        self.time_to = Timestamp(diff_to)

        super().__init__(order_by=["a.name"], **kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        self.params["ids"] = self.ids

        rels_filter, rels_params = self.branch.get_query_filter_relationships_range(
            rel_labels=["r"], start_time=self.time_from, end_time=self.time_to, include_outside_parentheses=True
        )

        self.params.update(rels_params)

        query = """
        MATCH (a) WHERE a.uuid IN $ids
        MATCH (a)-[r:IS_VISIBLE|IS_PROTECTED|HAS_SOURCE|HAS_OWNER|HAS_VALUE]-(ap)
        WHERE %s
        """ % ("\n AND ".join(rels_filter),)

        self.add_to_query(query)
        self.return_labels = ["a", "ap", "r"]

    def get_results_by_id_and_prop_type(self, attr_id: str, prop_type: str) -> list[QueryResult]:
        """Return a list of all results matching a given relationship id / property type.

        The results are ordered chronologicall (from oldest to newest)
        """
        results = [
            result
            for result in self.results
            if result.get("r").get("branch") in self.branch.get_branches_in_scope()
            and result.get("a").get("uuid") == attr_id
            and result.get("r").type == prop_type
        ]

        return sort_results_by_time(results, rel_label="r")


class DiffNodePropertiesByIDSQuery(Query):
    name = "diff_node_properties_ids"
    type = QueryType.READ
    order_by: list[str] = ["a.name"]

    def __init__(
        self,
        ids: list[str],
        at: str,
        account=None,
        **kwargs,
    ):
        self.account = account
        self.ids = ids
        self.at = Timestamp(at)

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        self.params["ids"] = self.ids

        rels_filter, rels_params = self.branch.get_query_filter_relationships(
            rel_labels=["r"], at=self.at, include_outside_parentheses=True
        )

        self.params.update(rels_params)

        query = """
        MATCH (a:Attribute) WHERE a.uuid IN $ids
        MATCH (a)-[r:IS_VISIBLE|IS_PROTECTED|HAS_SOURCE|HAS_OWNER|HAS_VALUE]-(ap)
        WHERE %s
        """ % ("\n AND ".join(rels_filter),)

        self.add_to_query(query)
        self.return_labels = ["a", "ap", "r"]

    def get_results_by_id_and_prop_type(self, attr_id: str, prop_type: str) -> list[QueryResult]:
        """Return a list of all results matching a given relationship id / property type.

        The results are ordered chronologicall (from oldest to newest)
        """
        results = [
            result
            for result in self.results
            if result.get("r").get("branch") in self.branch.get_branches_in_scope()
            and result.get("a").get("uuid") == attr_id
            and result.get("r").type == prop_type
        ]

        return sort_results_by_time(results, rel_label="r")


class DiffRelationshipPropertiesByIDSRangeQuery(Query):
    name = "diff_relationship_properties_range_ids"
    type = QueryType.READ

    def __init__(
        self,
        ids: list[str],
        diff_from: str,
        diff_to: str,
        account=None,
        **kwargs,
    ):
        self.account = account
        self.ids = ids
        self.time_from = Timestamp(diff_from)
        self.time_to = Timestamp(diff_to)

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        self.params["ids"] = self.ids

        rels_filter, rels_params = self.branch.get_query_filter_relationships_range(
            rel_labels=["r"], start_time=self.time_from, end_time=self.time_to, include_outside_parentheses=True
        )

        self.params.update(rels_params)

        # TODO Compute the list of potential relationship dynamically in the future based on the class
        query = """
        MATCH (rl:Relationship) WHERE rl.uuid IN $ids
        MATCH (rl)-[r:IS_VISIBLE|IS_PROTECTED|HAS_SOURCE|HAS_OWNER]-(rp)
        WHERE %s
        """ % ("\n AND ".join(rels_filter),)

        self.params["at"] = self.at.to_string()

        self.add_to_query(query)
        self.return_labels = ["rl", "rp", "r"]

    def get_results_by_id_and_prop_type(self, rel_id: str, prop_type: str) -> list[QueryResult]:
        """Return a list of all results matching a given relationship id / property type.
        The results are ordered chronologically
        """
        results = [
            result
            for result in self.results
            if result.get("r").get("branch") in self.branch.get_branches_in_scope()
            and result.get("rl").get("uuid") == rel_id
            and result.get("r").type == prop_type
        ]

        return sort_results_by_time(results, rel_label="r")


class DiffCountChanges(Query):
    name = "diff_count_changes"
    type = QueryType.READ

    def __init__(
        self,
        branch_names: list[str],
        diff_from: Timestamp,
        diff_to: Timestamp,
        **kwargs,
    ):
        self.branch_names = branch_names
        self.diff_from = diff_from
        self.diff_to = diff_to
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        self.params = {
            "from_time": self.diff_from.to_string(),
            "to_time": self.diff_to.to_string(),
            "branch_names": self.branch_names,
        }
        query = """
        MATCH (p)-[diff_rel]-(q)
        WHERE any(l in labels(p) WHERE l in ["Node", "Attribute", "Relationship"])
        AND diff_rel.branch in $branch_names
        AND (
            (diff_rel.from >= $from_time AND diff_rel.from < $to_time)
            OR (diff_rel.to >= $to_time AND diff_rel.to < $to_time)
        )
        AND (p.branch_support = "aware" OR q.branch_support = "aware")
        WITH diff_rel.branch AS branch_name, count(*) AS num_changes
        """
        self.add_to_query(query=query)
        self.return_labels = ["branch_name", "num_changes"]

    def get_num_changes_by_branch(self) -> dict[str, int]:
        branch_count_map = {}
        for result in self.get_results():
            branch_name = str(result.get("branch_name"))
            try:
                count = int(result.get("num_changes"))
            except (TypeError, ValueError):
                count = 0
            branch_count_map[branch_name] = count
        for branch_name in self.branch_names:
            if branch_name not in branch_count_map:
                branch_count_map[branch_name] = 0
        return branch_count_map


class DiffAllPathsQuery(DiffQuery):
    """Gets the required Cypher paths for a diff"""

    name = "diff_node"
    type = QueryType.READ

    def __init__(
        self,
        base_branch: Branch,
        diff_branch_from_time: Timestamp,
        current_node_field_specifiers: list[tuple[str, str]] | None = None,
        new_node_field_specifiers: list[tuple[str, str]] | None = None,
        **kwargs: Any,
    ):
        self.base_branch = base_branch
        self.diff_branch_from_time = diff_branch_from_time
        self.current_node_field_specifiers = current_node_field_specifiers
        self.new_node_field_specifiers = new_node_field_specifiers

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        from_str = self.diff_from.to_string()
        self.params.update(
            {
                "base_branch_name": self.base_branch.name,
                "branch_name": self.branch.name,
                "global_branch_name": GLOBAL_BRANCH_NAME,
                "branch_from_time": self.diff_branch_from_time.to_string(),
                "from_time": from_str,
                "to_time": self.diff_to.to_string(),
                "branch_local": BranchSupportType.LOCAL.value,
                "branch_aware": BranchSupportType.AWARE.value,
                "branch_agnostic": BranchSupportType.AGNOSTIC.value,
                "new_node_field_specifiers": self.new_node_field_specifiers,
                "current_node_field_specifiers": self.current_node_field_specifiers,
            }
        )
        query = """
WITH CASE
    WHEN $new_node_field_specifiers IS NULL AND $current_node_field_specifiers IS NULL THEN [[NULL, $from_time]]
    WHEN $new_node_field_specifiers IS NULL OR size($new_node_field_specifiers) = 0 THEN [[$current_node_field_specifiers, $from_time]]
    WHEN $current_node_field_specifiers IS NULL OR size($current_node_field_specifiers) = 0 THEN [[$new_node_field_specifiers, $branch_from_time]]
    ELSE [[$new_node_field_specifiers, $branch_from_time], [$current_node_field_specifiers, $from_time]]
END AS diff_filter_params_list
UNWIND diff_filter_params_list AS diff_filter_params
WITH diff_filter_params[0] AS node_field_specifiers_list, diff_filter_params[1] AS from_time
CALL {
    // -------------------------------------
    // These lists contain duplicate data, but vastly improve querying speed below
    // -------------------------------------
    WITH node_field_specifiers_list
    UNWIND node_field_specifiers_list AS nfs
    WITH nfs[0] AS uuid, nfs[1] AS field_name
    WITH collect(DISTINCT uuid) as uuids, collect(DISTINCT field_name) AS field_names
    RETURN uuids AS node_ids_list, field_names AS field_names_list
}
CALL {
    WITH node_field_specifiers_list, node_ids_list, from_time
    CALL {
        WITH node_field_specifiers_list, from_time, node_ids_list
        // -------------------------------------
        // Identify nodes added/removed on branch
        // -------------------------------------
        MATCH (q:Root)<-[diff_rel:IS_PART_OF {branch: $branch_name}]-(p:Node)
        WHERE (size(node_ids_list) = 0 OR p.uuid IN node_ids_list)
        AND (from_time <= diff_rel.from < $to_time)
        AND (diff_rel.to IS NULL OR (from_time <= diff_rel.to < $to_time))
        AND p.branch_support = $branch_aware
        WITH p, q, diff_rel, from_time
        // -------------------------------------
        // Exclude nodes added then removed on branch within timeframe
        // -------------------------------------
        CALL {
            WITH p, q, from_time
            OPTIONAL MATCH (q)<-[is_part_of:IS_PART_OF {branch: $branch_name}]-(p)
            WHERE from_time <= is_part_of.from < $to_time
            WITH DISTINCT is_part_of.status AS rel_status
            WITH collect(rel_status) AS rel_statuses
            RETURN ("active" IN rel_statuses AND "deleted" IN rel_statuses) AS intra_branch_update
        }
        WITH p, q, diff_rel, from_time, intra_branch_update
        WHERE intra_branch_update = FALSE
        // -------------------------------------
        // Get every path on this branch under each node
        // -------------------------------------
        CALL {
            WITH p, q, diff_rel, from_time
            OPTIONAL MATCH path = (
                (q)<-[top_diff_rel:IS_PART_OF]-(p)-[r_node]-(node)-[r_prop]-(prop)
            )
            WHERE %(id_func)s(diff_rel) = %(id_func)s(top_diff_rel)
            AND type(r_node) IN ["HAS_ATTRIBUTE", "IS_RELATED"]
            AND any(l in labels(node) WHERE l in ["Attribute", "Relationship"])
            AND node.branch_support IN [$branch_aware, $branch_agnostic]
            AND type(r_prop) IN ["IS_VISIBLE", "IS_PROTECTED", "HAS_SOURCE", "HAS_OWNER", "HAS_VALUE", "IS_RELATED"]
            AND any(l in labels(prop) WHERE l in ["Boolean", "Node", "AttributeValue"])
            AND ALL(
                r in [r_node, r_prop]
                WHERE r.from < $to_time AND r.branch = top_diff_rel.branch
            )
            AND (top_diff_rel.to IS NULL OR top_diff_rel.to >= r_node.from)
            AND (r_node.to IS NULL OR r_node.to >= r_prop.from)
            AND [%(id_func)s(p), type(r_node)] <> [%(id_func)s(prop), type(r_prop)]
            AND top_diff_rel.status = r_node.status
            AND top_diff_rel.status = r_prop.status
            WITH path, p, node, prop, r_prop, r_node, type(r_node) AS rel_type, from_time
            // -------------------------------------
            // Exclude attributes/relationships added then removed on branch within timeframe
            // -------------------------------------
            CALL {
                WITH p, rel_type, node, from_time
                OPTIONAL MATCH (p)-[rel_to_check {branch: $branch_name}]-(node)
                WHERE from_time <= rel_to_check.from < $to_time
                AND type(rel_to_check) = rel_type
                WITH DISTINCT rel_to_check.status AS rel_status
                WITH collect(rel_status) AS rel_statuses
                RETURN ("active" IN rel_statuses AND "deleted" IN rel_statuses) AS intra_branch_update
            }
            WITH path, node, prop, r_prop, r_node, intra_branch_update
            WHERE intra_branch_update = FALSE
            WITH path, node, prop, r_prop, r_node
            ORDER BY
                %(id_func)s(node),
                %(id_func)s(prop),
                r_prop.from DESC,
                r_node.from DESC
            WITH node, prop, type(r_prop) AS r_prop_type, type(r_node) AS r_node_type, head(collect(path)) AS top_diff_path
            RETURN top_diff_path
        }
        RETURN top_diff_path AS diff_path
    }
    RETURN diff_path
    UNION
    WITH node_field_specifiers_list, node_ids_list, field_names_list, from_time
    CALL {
        WITH node_field_specifiers_list, node_ids_list, field_names_list, from_time
        // -------------------------------------
        // Identify attributes/relationships added/removed on branch
        // -------------------------------------
        CALL {
            WITH node_field_specifiers_list, node_ids_list, field_names_list, from_time
            MATCH (root:Root)<-[r_root:IS_PART_OF]-(p:Node)-[diff_rel:HAS_ATTRIBUTE {branch: $branch_name}]->(q:Attribute)
            // exclude attributes and relationships under added/removed nodes b/c they are covered above
            WHERE r_root.branch IN [$branch_name, $base_branch_name, $global_branch_name]
            AND q.branch_support = $branch_aware
            AND (size(node_ids_list) = 0 OR p.uuid IN node_ids_list)
            AND (size(field_names_list) = 0 OR q.name IN field_names_list)
            // if p has a different type of branch support and was addded within our timeframe
            AND (r_root.from < from_time OR p.branch_support = $branch_agnostic)
            AND r_root.status = "active"
            // get attributes and relationships added on the branch during the timeframe
            AND (from_time <= diff_rel.from < $to_time)
            AND (diff_rel.to IS NULL OR (from_time <= diff_rel.to < $to_time))
            AND r_root.from <= diff_rel.from
            AND (r_root.to IS NULL OR diff_rel.branch <> r_root.branch OR r_root.to >= diff_rel.from)
            AND (node_field_specifiers_list IS NULL OR [p.uuid, q.name] IN node_field_specifiers_list)
            RETURN root, r_root, p, diff_rel, q
            UNION ALL
            WITH node_field_specifiers_list, node_ids_list, field_names_list, from_time
            MATCH (root:Root)<-[r_root:IS_PART_OF]-(p:Node)-[diff_rel:IS_RELATED {branch: $branch_name}]-(q:Relationship)
            // exclude attributes and relationships under added/removed nodes b/c they are covered above
            WHERE r_root.branch IN [$branch_name, $base_branch_name, $global_branch_name]
            AND q.branch_support = $branch_aware
            AND (size(node_ids_list) = 0 OR p.uuid IN node_ids_list)
            AND (size(field_names_list) = 0 OR q.name IN field_names_list)
            // if p has a different type of branch support and was addded within our timeframe
            AND (r_root.from < from_time OR p.branch_support = $branch_agnostic)
            // get attributes and relationships added on the branch during the timeframe
            AND (from_time <= diff_rel.from < $to_time)
            AND (diff_rel.to IS NULL OR (from_time <= diff_rel.to < $to_time))
            AND r_root.from <= diff_rel.from
            AND (r_root.to IS NULL OR diff_rel.branch <> r_root.branch OR r_root.to >= diff_rel.from)
            AND (node_field_specifiers_list IS NULL OR [p.uuid, q.name] IN node_field_specifiers_list)
            RETURN root, r_root, p, diff_rel, q
        }
        WITH root, r_root, p, diff_rel, q, from_time
        // -------------------------------------
        // Exclude attributes/relationship under nodes deleted on this branch in the timeframe
        // because those were all handled above at the node level
        // -------------------------------------
        CALL {
            WITH root, p, from_time
            OPTIONAL MATCH (root)<-[r_root_deleted:IS_PART_OF {branch: $branch_name}]-(p)
            WHERE from_time <= r_root_deleted.from < $to_time
            WITH r_root_deleted
            ORDER BY r_root_deleted.status DESC
            LIMIT 1
            RETURN COALESCE(r_root_deleted.status = "deleted", FALSE) AS node_deleted
        }
        WITH root, r_root, p, diff_rel, q, from_time, node_deleted
        WHERE node_deleted = FALSE
        // -------------------------------------
        // Exclude relationships added and deleted within the timeframe
        // -------------------------------------
        WITH root, r_root, p, diff_rel, q, from_time, type(diff_rel) AS rel_type
        CALL {
            WITH p, rel_type, q, from_time
            OPTIONAL MATCH (p)-[rel_to_check {branch: $branch_name}]-(q)
            WHERE from_time <= rel_to_check.from < $to_time
            AND type(rel_to_check) = rel_type
            WITH DISTINCT rel_to_check.status AS rel_status
            WITH collect(rel_status) AS rel_statuses
            RETURN ("active" IN rel_statuses AND "deleted" IN rel_statuses) AS intra_branch_update
        }
        WITH root, r_root, p, diff_rel, q, from_time, intra_branch_update
        WHERE intra_branch_update = FALSE
        // -------------------------------------
        // Get every path on this branch under each attribute/relationship
        // -------------------------------------
        CALL {
            WITH root, r_root, p, diff_rel, q
            OPTIONAL MATCH path = (
                (root:Root)<-[mid_r_root:IS_PART_OF]-(p)-[mid_diff_rel]-(q)-[r_prop]-(prop)
            )
            WHERE %(id_func)s(mid_r_root) =  %(id_func)s(r_root)
            AND %(id_func)s(mid_diff_rel) =  %(id_func)s(diff_rel)
            AND type(r_prop) IN ["IS_VISIBLE", "IS_PROTECTED", "HAS_SOURCE", "HAS_OWNER", "HAS_VALUE", "IS_RELATED"]
            AND any(l in labels(prop) WHERE l in ["Boolean", "Node", "AttributeValue"])
            AND r_prop.from < $to_time AND r_prop.branch = mid_diff_rel.branch
            AND (mid_diff_rel.to IS NULL OR mid_diff_rel.to >= r_prop.from)
            AND [%(id_func)s(p), type(mid_diff_rel)] <> [%(id_func)s(prop), type(r_prop)]
            // exclude paths where an active edge is below a deleted edge
            AND (mid_diff_rel.status = "active" OR r_prop.status = "deleted")
            WITH path, prop, r_prop, mid_r_root
            ORDER BY
                type(r_prop),
                mid_r_root.branch = mid_diff_rel.branch DESC,
                r_prop.from DESC,
                mid_r_root.from DESC
            WITH prop, type(r_prop) AS type_r_prop, head(collect(path)) AS latest_prop_path
            RETURN latest_prop_path
        }
        // -------------------------------------
        // Exclude properties within the timeframe
        // -------------------------------------
        WITH q, nodes(latest_prop_path)[3] AS prop, type(relationships(latest_prop_path)[2]) AS rel_type, latest_prop_path, from_time
        CALL {
            WITH q, rel_type, prop, from_time
            OPTIONAL MATCH (q)-[rel_to_check {branch: $branch_name}]-(prop)
            WHERE from_time <= rel_to_check.from < $to_time
            AND type(rel_to_check) = rel_type
            WITH DISTINCT rel_to_check.status AS rel_status
            WITH collect(rel_status) AS rel_statuses
            RETURN ("active" IN rel_statuses AND "deleted" IN rel_statuses) AS intra_branch_update
        }
        WITH latest_prop_path, intra_branch_update
        WHERE intra_branch_update = FALSE
        RETURN latest_prop_path AS mid_diff_path
    }
    RETURN mid_diff_path AS diff_path
    UNION
    WITH node_field_specifiers_list, node_ids_list, field_names_list, from_time
    CALL {
        WITH node_field_specifiers_list, node_ids_list, field_names_list, from_time
        // -------------------------------------
        // Identify properties added/removed on branch
        // -------------------------------------
        MATCH diff_rel_path = (root:Root)<-[r_root:IS_PART_OF]-(n:Node)-[r_node]-(p)-[diff_rel {branch: $branch_name}]->(q)
        WHERE (
            (from_time <= diff_rel.from < $to_time)
            OR (from_time <= diff_rel.to < $to_time)
        )
        AND (size(node_ids_list) = 0 OR n.uuid IN node_ids_list)
        AND (size(field_names_list) = 0 OR p.name IN field_names_list)
        // exclude attributes and relationships under added/removed nodes, attrs, and rels b/c they are covered above
        AND ALL(
            r in [r_root, r_node]
            WHERE r.from <= from_time AND r.branch IN [$branch_name, $base_branch_name]
        )
        AND p.branch_support = $branch_aware
        AND any(l in labels(p) WHERE l in ["Attribute", "Relationship"])
        AND type(diff_rel) IN ["IS_VISIBLE", "IS_PROTECTED", "HAS_SOURCE", "HAS_OWNER", "HAS_VALUE"]
        AND any(l in labels(q) WHERE l in ["Boolean", "Node", "AttributeValue"])
        AND type(r_node) IN ["HAS_ATTRIBUTE", "IS_RELATED"]
        AND (node_field_specifiers_list IS NULL OR [n.uuid, p.name] IN node_field_specifiers_list)
        AND ALL(
            r_pair IN [[r_root, r_node], [r_node, diff_rel]]
            // filter out paths where a base branch edge follows a branch edge
            WHERE ((r_pair[0]).branch = $base_branch_name OR (r_pair[1]).branch = $branch_name)
            // filter out paths where an active edge follows a deleted edge
            AND ((r_pair[0]).status = "active" OR (r_pair[1]).status = "deleted")
            // filter out paths where an earlier from time follows a later from time
            AND (r_pair[0]).from <= (r_pair[1]).from
            // require adjacent edge pairs to have overlapping times, but only if on the same branch
            AND (
                (r_pair[0]).branch <> (r_pair[1]).branch
                OR (r_pair[0]).to IS NULL
                OR (r_pair[0]).to >= (r_pair[1]).from
            )
        )
        AND [%(id_func)s(n), type(r_node)] <> [%(id_func)s(q), type(diff_rel)]
        WITH diff_rel_path, r_root, n, r_node, p, diff_rel, from_time
        ORDER BY
            %(id_func)s(n) DESC,
            %(id_func)s(p) DESC,
            type(diff_rel),
            r_node.branch = diff_rel.branch DESC,
            r_root.branch = diff_rel.branch DESC,
            diff_rel.from DESC,
            r_node.from DESC,
            r_root.from DESC
        WITH n, p, from_time, diff_rel, diff_rel_path
        CALL {
            // -------------------------------------
            // Exclude properties under nodes and attributes/relationship deleted
            // on this branch in the timeframe because those were all handled above
            // -------------------------------------
            WITH n, p, from_time
            CALL {
                WITH n, from_time
                OPTIONAL MATCH (root:Root)<-[r_root_deleted:IS_PART_OF {branch: $branch_name}]-(n)
                WHERE from_time <= r_root_deleted.from < $to_time
                WITH r_root_deleted
                ORDER BY r_root_deleted.status DESC
                LIMIT 1
                RETURN COALESCE(r_root_deleted.status = "deleted", FALSE) AS node_deleted
            }
            WITH n, p, from_time, node_deleted
            CALL {
                WITH n, p, from_time
                OPTIONAL MATCH (n)-[r_node_deleted {branch: $branch_name}]-(p)
                WHERE from_time <= r_node_deleted.from < $to_time
                AND type(r_node_deleted) IN ["HAS_ATTRIBUTE", "IS_RELATED"]
                WITH r_node_deleted
                ORDER BY r_node_deleted.status DESC
                LIMIT 1
                RETURN COALESCE(r_node_deleted.status = "deleted", FALSE) AS field_deleted
            }
            RETURN node_deleted OR field_deleted AS node_or_field_deleted
        }
        WITH n, p, diff_rel, diff_rel_path, node_or_field_deleted
        WHERE node_or_field_deleted = FALSE
        WITH n, p, type(diff_rel) AS drt, head(collect(diff_rel_path)) AS deepest_diff_path
        RETURN deepest_diff_path
    }
    RETURN deepest_diff_path AS diff_path
}
WITH DISTINCT diff_path AS diff_path
CALL {
    WITH diff_path
    WITH diff_path, nodes(diff_path) AS d_nodes, relationships(diff_path) AS d_rels
    WITH diff_path, d_rels[0] AS r_root, d_nodes[1] AS n, d_rels[1] AS r_node, d_nodes[2] AS attr_rel, d_rels[2] AS r_prop
    // -------------------------------------
    // add base branch paths before branched_from, if they exist
    // -------------------------------------
    WITH n, attr_rel, r_node, r_prop
    OPTIONAL MATCH latest_base_path = (:Root)<-[base_r_root:IS_PART_OF {branch: $base_branch_name}]
        -(n)-[base_r_node {branch: $base_branch_name}]
        -(attr_rel)-[base_r_prop {branch: $base_branch_name}]->(base_prop)
    WHERE type(base_r_node) = type(r_node)
    AND type(base_r_prop) = type(r_prop)
    AND [%(id_func)s(n), type(base_r_node)] <> [%(id_func)s(base_prop), type(base_r_prop)]
    AND all(
        r in relationships(latest_base_path)
        WHERE r.from < $branch_from_time
    )
    WITH latest_base_path, base_r_root, base_r_node, base_r_prop
    ORDER BY base_r_prop.from DESC, base_r_node.from DESC, base_r_root.from DESC
    LIMIT 1
    RETURN latest_base_path
}
WITH diff_path, latest_base_path
UNWIND [diff_path, latest_base_path] AS penultimate_path
WITH penultimate_path
CALL {
    WITH penultimate_path
    WITH penultimate_path, nodes(penultimate_path) AS d_nodes, relationships(penultimate_path) AS d_rels
    WITH penultimate_path, d_rels[0] AS r_root, d_nodes[1] AS n, d_rels[1] AS r_node, d_nodes[2] AS attr_rel, d_rels[2] AS r_prop
    // -------------------------------------
    // Add peer-side of any relationships to get the peer's ID
    // -------------------------------------
    WITH r_root, n, r_node, attr_rel, r_prop
    OPTIONAL MATCH peer_path = (
        (:Root)<-[peer_r_root:IS_PART_OF]-(n)-[peer_r_node:IS_RELATED]-(attr_rel:Relationship)-[r_peer:IS_RELATED]-(peer:Node)
    )
    WHERE type(r_prop) <> "IS_RELATED"
    AND %(id_func)s(peer_r_root) = %(id_func)s(r_root)
    AND %(id_func)s(peer_r_node) = %(id_func)s(r_node)
    AND [%(id_func)s(n), type(peer_r_node)] <> [%(id_func)s(peer), type(r_peer)]
    AND r_peer.from < $to_time
    // filter out paths where an earlier from time follows a later from time
    AND peer_r_node.from <= r_peer.from
    // filter out paths where a base branch edge follows a branch edge
    AND (peer_r_node.branch = $base_branch_name OR r_peer.branch = $branch_name)
    // filter out paths where an active edge follows a deleted edge
    AND (peer_r_node.status = "active" OR r_peer.status = "deleted")
    // require adjacent edge pairs to have overlapping times, but only if on the same branch
    AND (
        peer_r_node.branch <> r_peer.branch
        OR peer_r_node.to IS NULL
        OR peer_r_node.to >= r_peer.from
    )
    WITH peer_path, r_peer, r_prop
    ORDER BY r_peer.branch = r_prop.branch DESC, r_peer.from DESC
    LIMIT 1
    RETURN peer_path
}
WITH penultimate_path, peer_path
WITH reduce(
    diff_rel_paths = [], item IN [penultimate_path, peer_path] |
    CASE WHEN item IS NULL THEN diff_rel_paths ELSE diff_rel_paths + [item] END
) AS diff_rel_paths
UNWIND diff_rel_paths AS diff_path
        """ % {"id_func": db.get_id_function_name()}
        self.add_to_query(query)
        self.return_labels = ["DISTINCT diff_path AS diff_path"]



from collections import Counter
from typing import Iterable

from infrahub.core.constants import DiffAction

from ..model.path import (
    BaseSummary,
    CalculatedDiffs,
    EnrichedDiffAttribute,
    EnrichedDiffNode,
    EnrichedDiffRelationship,
    EnrichedDiffRoot,
    EnrichedDiffSingleRelationship,
)
from .interface import DiffEnricherInterface


class DiffSummaryCountsEnricher(DiffEnricherInterface):
    async def enrich(
        self, enriched_diff_root: EnrichedDiffRoot, calculated_diffs: CalculatedDiffs | None = None
    ) -> None:
        self._add_root_summaries(diff_root=enriched_diff_root)

    def _add_summary(self, summary_node: BaseSummary, actions: Iterable[DiffAction]) -> None:
        summary_count = Counter(actions)
        summary_node.num_added = summary_count.get(DiffAction.ADDED, 0)
        summary_node.num_updated = summary_count.get(DiffAction.UPDATED, 0)
        summary_node.num_removed = summary_count.get(DiffAction.REMOVED, 0)

    def _add_root_summaries(self, diff_root: EnrichedDiffRoot) -> None:
        contains_conflict = False
        num_conflicts = 0
        for diff_node in diff_root.nodes:
            contains_conflict |= self._add_node_summaries(diff_node=diff_node)
            if diff_node.conflict or diff_node.contains_conflict:
                num_conflicts += 1
        self._add_summary(summary_node=diff_root, actions=(n.action for n in diff_root.nodes))
        diff_root.contains_conflict = contains_conflict
        diff_root.num_conflicts = num_conflicts
        self._add_child_nodes_to_summaries(diff_root=diff_root)

    def _add_node_summaries(self, diff_node: EnrichedDiffNode) -> bool:
        contains_conflict = False
        num_conflicts = 0
        for diff_attr in diff_node.attributes:
            contains_conflict |= self._add_attribute_summaries(diff_attribute=diff_attr)
            if diff_attr.contains_conflict:
                num_conflicts += 1
        for diff_rel in diff_node.relationships:
            contains_conflict |= self._add_relationship_summaries(diff_relationship=diff_rel)
            if diff_rel.contains_conflict:
                num_conflicts += 1
        self._add_summary(
            summary_node=diff_node, actions=(field.action for field in diff_node.relationships | diff_node.attributes)
        )
        diff_node.contains_conflict = contains_conflict
        diff_node.num_conflicts = num_conflicts
        return contains_conflict

    def _add_attribute_summaries(self, diff_attribute: EnrichedDiffAttribute) -> bool:
        contains_conflict = False
        num_conflicts = 0
        for diff_prop in diff_attribute.properties:
            if diff_prop.conflict:
                num_conflicts += 1
                contains_conflict = True
        self._add_summary(summary_node=diff_attribute, actions=(p.action for p in diff_attribute.properties))
        diff_attribute.contains_conflict = contains_conflict
        diff_attribute.num_conflicts = num_conflicts
        return contains_conflict

    def _add_relationship_summaries(self, diff_relationship: EnrichedDiffRelationship) -> bool:
        contains_conflict = False
        num_conflicts = 0
        for diff_element in diff_relationship.relationships:
            contains_conflict |= self._add_element_summaries(diff_element=diff_element)
            if diff_element.conflict:
                num_conflicts += 1
        self._add_summary(summary_node=diff_relationship, actions=(e.action for e in diff_relationship.relationships))
        diff_relationship.contains_conflict = contains_conflict
        diff_relationship.num_conflicts = num_conflicts
        return contains_conflict

    def _add_element_summaries(self, diff_element: EnrichedDiffSingleRelationship) -> bool:
        if diff_element.conflict is None:
            contains_conflict = False
            num_conflicts = 0
        else:
            contains_conflict = True
            num_conflicts = 1
        for diff_prop in diff_element.properties:
            if diff_prop.conflict:
                num_conflicts += 1
                contains_conflict = True
        self._add_summary(summary_node=diff_element, actions=(p.action for p in diff_element.properties))
        diff_element.contains_conflict = contains_conflict
        diff_element.num_conflicts = num_conflicts
        return contains_conflict

    def _add_child_nodes_to_summaries(self, diff_root: EnrichedDiffRoot) -> None:
        for diff_node in diff_root.nodes:
            for diff_rel in diff_node.relationships:
                if not diff_rel.contains_conflict:
                    diff_rel.contains_conflict = any(n.contains_conflict for n in diff_rel.nodes)
                diff_rel.num_conflicts += sum(bool(n.contains_conflict or n.conflict) for n in diff_rel.nodes)



from infrahub.core.validators.attribute.regex import AttributeRegexChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaAttributeRegexConstraintDependency(DependencyBuilder[AttributeRegexChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> AttributeRegexChecker:
        return AttributeRegexChecker(db=context.db, branch=context.branch)



from __future__ import annotations

from graphene import InputObjectType, String


class IdentifierInput(InputObjectType):
    id = String(required=True, description="The ID of the requested object")



from typing import Any, Optional, Protocol


class InfrahubLogger(Protocol):
    def debug(self, event: Optional[str] = None, *args: Any, **kw: Any) -> Any:
        """Send a debug event"""

    def info(self, event: Optional[str] = None, *args: Any, **kw: Any) -> Any:
        """Send an info event"""

    def warning(self, event: Optional[str] = None, *args: Any, **kw: Any) -> Any:
        """Send a warning event"""

    def error(self, event: Optional[str] = None, *args: Any, **kw: Any) -> Any:
        """Send an error event."""

    def critical(self, event: Optional[str] = None, *args: Any, **kw: Any) -> Any:
        """Send a critical event."""

    def exception(self, event: Optional[str] = None, *args: Any, **kw: Any) -> Any:
        """Send an exception event."""



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType
from infrahub.core.root import Root

from ..shared import GraphMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class Migration001Query01(Query):
    name = "migration_001_01"
    type: QueryType = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        query = """
        MATCH (root:Root)
        SET root.graph_version = 1
        """
        self.add_to_query(query)
        self.return_labels = ["root"]


class Migration001(GraphMigration):
    name: str = "001_add_version_to_graph"
    queries: Sequence[type[Query]] = [Migration001Query01]
    minimum_version: int = 0

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        result = MigrationResult()

        roots = await Root.get_list(db=db)
        if len(roots) == 0:
            result.errors.append("Unable to find the root node")
        elif len(roots) > 1:
            result.errors.append("Database is corrupted, more than 1 root node found.")

        if result.errors:
            return result

        root_node = roots[0]
        if root_node.graph_version != 1:
            result.errors.append(f"The version of the graph should be 1, found {root_node.graph_version!r} instead.")

        return result



from __future__ import annotations

from typing import Any, Union

from infrahub.core.path import GroupedDataPaths, SchemaPath  # noqa: TC001
from infrahub.core.query import Query, QueryType
from infrahub.core.schema import AttributeSchema, GenericSchema, NodeSchema, RelationshipSchema  # noqa: TC001


class SchemaValidatorQuery(Query):
    type: QueryType = QueryType.READ

    def __init__(
        self,
        node_schema: Union[NodeSchema, GenericSchema],
        schema_path: SchemaPath,
        **kwargs: Any,
    ) -> None:
        self.node_schema = node_schema
        self.schema_path = schema_path
        super().__init__(**kwargs)

    async def get_paths(self) -> GroupedDataPaths:
        raise NotImplementedError()


class AttributeSchemaValidatorQuery(SchemaValidatorQuery):
    @property
    def attribute_schema(self) -> AttributeSchema:
        if not self.schema_path.field_name:
            raise ValueError("field_name not defined")
        return self.node_schema.get_attribute(name=self.schema_path.field_name)


class RelationshipSchemaValidatorQuery(SchemaValidatorQuery):
    @property
    def relationship_schema(self) -> RelationshipSchema:
        if not self.schema_path.field_name:
            raise ValueError("field_name not defined")
        return self.node_schema.get_relationship(name=self.schema_path.field_name)



from __future__ import annotations

from typing import TYPE_CHECKING, Union

from graphene.types.definitions import GrapheneInterfaceType, GrapheneObjectType
from graphql import (  # pylint: disable=no-name-in-module
    ExecutionContext,
    FieldNode,
    FragmentSpreadNode,
    GraphQLList,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLSchema,
    GraphQLUnionType,
    InlineFragmentNode,
    SelectionSetNode,
)

from infrahub.exceptions import GraphQLQueryError

if TYPE_CHECKING:
    import abc

    from graphql.execution import ExecutionResult


def extract_data(query_name: str, result: ExecutionResult) -> dict:
    if result.errors:
        errors = []
        for error in result.errors:
            error_locations = error.locations or []
            errors.append(
                {
                    "message": f"GraphQLQuery {query_name}: {error.message}",
                    "path": error.path,
                    "locations": [{"line": location.line, "column": location.column} for location in error_locations],
                }
            )

        raise GraphQLQueryError(errors=errors)

    return result.data or {}


# --------------------------------------------------------------
# The functions below :
#   - selected_field_names_fast
#   - selected_field_names_naive
#   - selected_field_names
#   - selected_field_names_from_context
# Are not currently used and they have been copied from internet as a reference
#  >>  https://github.com/graphql-python/graphene/issues/57#issuecomment-774227086
#
# --------------------------------------------------------------
def selected_field_names_fast(
    selection_set: SelectionSetNode, context: GraphQLResolveInfo, runtime_type: Union[str, GraphQLObjectType] = None
) -> abc.Iterator[str]:
    """Use the fastest available function to provide the list of selected field names

    Note that this function may give false positives because in the absence of fragments it ignores directives.
    """
    # Any fragments?
    no_fragments = all(isinstance(node, FieldNode) for node in selection_set.selections)

    # Choose the function to execute
    if no_fragments:
        return selected_field_names_naive(selection_set)

    return selected_field_names(selection_set, context, runtime_type)


def selected_field_names_naive(selection_set: SelectionSetNode) -> list:
    """Get the list of field names that are selected at the current level. Does not include nested names.

    Limitations:
    * Does not resolve fragments; throws RuntimeError
    * Does not take directives into account. A field might be disabled, and this function wouldn't know

    As a result:
    * It will give a RuntimeError if a fragment is provided
    * It may give false positives in case directives are used
    * It is 20x faster than the alternative

    Benefits:
    * Fast!

    Args:
        selection_set: the selected fields

    Code copied from https://github.com/graphql-python/graphene/issues/57#issuecomment-774227086
        This link also includes a more powerful and complexe alternative.
    """
    assert isinstance(selection_set, SelectionSetNode)

    field_names = []

    for node in selection_set.selections:
        # Field
        if isinstance(node, FieldNode):
            field_names.append(node.name.value)
        # Fragment spread (`... fragmentName`)
        elif isinstance(node, (FragmentSpreadNode, InlineFragmentNode)):
            raise NotImplementedError("Fragments are not supported by this simplistic function")
        # Something new
        else:
            raise NotImplementedError(str(type(node)))

    return field_names


def selected_field_names(
    selection_set: SelectionSetNode, info: GraphQLResolveInfo, runtime_type: Union[str, GraphQLObjectType] = None
) -> abc.Iterator[str]:
    """Get the list of field names that are selected at the current level. Does not include nested names.

    This function re-evaluates the AST, but gives a complete list of included fields.
    It is 25x slower than `selected_field_names_naive()`, but still, it completes in 7ns or so. Not bad.

    Args:
        selection_set: the selected fields
        info: GraphQL resolve info
        runtime_type: The type of the object you resolve to. Either its string name, or its ObjectType.
            If none is provided, this function will fail with a RuntimeError() when resolving fragments
    """
    # pylint: disable=no-value-for-parameter

    # Create a temporary execution context. This operation is quite cheap, actually.
    execution_context = ExecutionContext(
        schema=info.schema,
        fragments=info.fragments,
        root_value=info.root_value,
        operation=info.operation,
        variable_values=info.variable_values,
        # The only purpose of this context is to be able to run the collect_fields() method.
        # Therefore, many parameters are actually irrelevant
        context_value=None,
        field_resolver=None,
        type_resolver=None,
        errors=[],
        middleware_manager=None,
    )

    # Use it
    return selected_field_names_from_context(selection_set, execution_context, runtime_type)


def selected_field_names_from_context(
    selection_set: SelectionSetNode, context: ExecutionContext, runtime_type: Union[str, GraphQLObjectType] = None
) -> abc.Iterator[str]:
    """Get the list of field names that are selected at the current level.

    This function is useless because `graphql.ExecutionContext` is not available at all inside resolvers.
    Therefore, `selected_field_names()` wraps it and provides one.
    """
    assert isinstance(selection_set, SelectionSetNode)

    # Resolve `runtime_type`
    if isinstance(runtime_type, str):
        runtime_type = context.schema.type_map[runtime_type]  # raises: KeyError

    # Resolve all fields
    fields_map = context.collect_fields(
        # Use the provided Object type, or use a dummy object that fails all tests
        runtime_type=runtime_type or None,
        # runtime_type=runtime_type or graphql.GraphQLObjectType('<temp>', []),
        selection_set=selection_set,
        fields={},  # out
        visited_fragment_names=(visited_fragment_names := set()),  # out
    )

    # Test fragment resolution
    if visited_fragment_names and not runtime_type:
        raise RuntimeError(
            "The query contains fragments which cannot be resolved "
            "because `runtime_type` is not provided by the lazy developer"
        )

    # Results!
    return (field.name.value for fields_list in fields_map.values() for field in fields_list)


def print_query(info: GraphQLResolveInfo) -> None:
    """Traverse the query"""
    initial_selection_set = info.field_nodes[0].selection_set
    print_selection_set(initial_selection_set, 1)


def print_selection_set(selection_set: SelectionSetNode, level: int = 1) -> int:
    # max_depth = level
    tab = "  "
    for field in getattr(selection_set, "selections", []):
        # print(f"in print_selection_set loop {field}")
        # The field we are at is already a lever deeper, even if it doesn't have its own selection set.
        # max_depth = max(max_depth, level + 1)
        print(f"{level * tab}{field.name.value}")
        if selection_set := getattr(field, "selection_set", None):
            # max_depth = max(max_depth, self._get_query_depth(selection_set, level + 1))
            print_selection_set(selection_set, level + 1)

    # """
    # MATCH (d:Device)
    # WHERE (d)-[]-(:Attribute {name: "name"})-[]-(:AttributeValue {value: "spine1"})
    # MAT36CH (d)-[:IS_RELATED]-(r1:Relationship{type: "device_interface"})-[:IS_RELATED]-(n1)
    # WHERE (n1)-[]-(:Attribute {name: "enabled"})-[]-(:AttributeValue {value: false})
    # OPTIONAL MATCH (d)-[:IS_RELATED]-(r1)-[:IS_RELATED]-(n1)-[:IS_RELATED]-(r2:Relationship {type: "interface_ip"})-[:IS_RELATED]-(n2)
    # RETURN d,n1,r1,n2,r2
    # """


def find_types_implementing_interface(
    interface: GrapheneInterfaceType, root_schema: GraphQLSchema
) -> list[GrapheneObjectType]:
    results = []
    for value in root_schema.type_map.values():
        if not hasattr(value, "interfaces"):
            continue

        for item in value.interfaces:
            if item.name == interface.name:
                results.append(value)

    return results


async def extract_schema_models(
    fields: dict, schema: Union[GrapheneObjectType, GraphQLUnionType], root_schema: GraphQLSchema
) -> set[str]:
    response = set()

    if isinstance(schema, GraphQLUnionType):
        return {t.name for t in schema.types}

    for field_name, value in fields.items():
        if field_name not in schema.fields:
            continue

        if isinstance(schema.fields[field_name].type, GrapheneObjectType):
            object_type = schema.fields[field_name].type
        elif isinstance(schema.fields[field_name].type, GraphQLList):
            object_type = schema.fields[field_name].type.of_type
        elif isinstance(schema.fields[field_name].type, GrapheneInterfaceType):
            object_type = schema.fields[field_name].type
            sub_types = find_types_implementing_interface(interface=object_type, root_schema=root_schema)
            for sub_type in sub_types:
                response.add(sub_type.name)
                response.update(await extract_schema_models(fields=value, schema=sub_type, root_schema=root_schema))
        else:
            continue

        response.add(object_type.name)

        if isinstance(value, dict):
            response.update(await extract_schema_models(fields=value, schema=object_type, root_schema=root_schema))
        elif isinstance(value, str) and value in schema.fields:
            if isinstance(schema.fields[value].type, GrapheneObjectType):
                response.add(schema.fields[value].type.name)
            elif isinstance(schema.fields[value].type, GraphQLList):
                response.add(schema.fields[value].type.of_type.name)

    return response



from . import (
    artifact_definition,
    generator_definition,
    proposed_change,
    repository,
)

__all__ = [
    "artifact_definition",
    "generator_definition",
    "proposed_change",
    "repository",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Iterable, Optional, Union

import graphene

from infrahub import config
from infrahub.core.attribute import String
from infrahub.core.constants import InfrahubKind, RelationshipKind
from infrahub.core.schema import (
    AttributeSchema,
    GenericSchema,
    MainSchemaTypes,
    NodeSchema,
    ProfileSchema,
    RelationshipSchema,
)
from infrahub.core.timestamp import Timestamp
from infrahub.graphql.mutations.attribute import BaseAttributeCreate, BaseAttributeUpdate
from infrahub.graphql.mutations.graphql_query import InfrahubGraphQLQueryMutation
from infrahub.types import ATTRIBUTE_TYPES, InfrahubDataType, get_attribute_type

from .directives import DIRECTIVES
from .enums import generate_graphql_enum, get_enum_attribute_type_name
from .metrics import SCHEMA_GENERATE_GRAPHQL_METRICS
from .mutations.artifact_definition import InfrahubArtifactDefinitionMutation
from .mutations.ipam import (
    InfrahubIPAddressMutation,
    InfrahubIPNamespaceMutation,
    InfrahubIPPrefixMutation,
)
from .mutations.main import InfrahubMutation
from .mutations.menu import InfrahubCoreMenuMutation
from .mutations.proposed_change import InfrahubProposedChangeMutation
from .mutations.repository import InfrahubRepositoryMutation
from .mutations.resource_manager import (
    InfrahubNumberPoolMutation,
)
from .resolvers.resolver import (
    account_resolver,
    ancestors_resolver,
    default_paginated_list_resolver,
    default_resolver,
    descendants_resolver,
    many_relationship_resolver,
    parent_field_name_resolver,
    single_relationship_resolver,
)
from .schema import InfrahubBaseMutation, InfrahubBaseQuery
from .subscription import InfrahubBaseSubscription
from .types import (
    InfrahubInterface,
    InfrahubObject,
    PaginatedObjectPermission,
    RelatedIPAddressNodeInput,
    RelatedNodeInput,
    RelatedPrefixNodeInput,
)
from .types.attribute import BaseAttribute as BaseAttributeType
from .types.attribute import TextAttributeType

if TYPE_CHECKING:
    from graphql import GraphQLSchema

    from infrahub.core.branch import Branch
    from infrahub.core.schema.schema_branch import SchemaBranch

# pylint: disable=redefined-builtin,c-extension-no-member,too-many-lines,too-many-public-methods


class DeleteInput(graphene.InputObjectType):
    id = graphene.String(required=False)
    hfid = graphene.List(of_type=graphene.String, required=False)


GraphQLTypes = Union[
    type[InfrahubMutation], type[BaseAttributeType], type[graphene.Interface], type[graphene.ObjectType]
]


class OrderInput(graphene.InputObjectType):
    disable = graphene.Boolean(required=False)


@dataclass
class GraphqlMutations:
    create: type[InfrahubMutation]
    update: type[InfrahubMutation]
    upsert: type[InfrahubMutation]
    delete: type[InfrahubMutation]


def get_attr_kind(node_schema: MainSchemaTypes, attr_schema: AttributeSchema) -> str:
    if not config.SETTINGS.experimental_features.graphql_enums or not attr_schema.enum:
        return attr_schema.kind
    return get_enum_attribute_type_name(node_schema=node_schema, attr_schema=attr_schema)


@dataclass
class BranchDetails:
    branch_name: str
    schema_changed_at: Timestamp
    schema_hash: str
    gql_manager: GraphQLSchemaManager


class GraphQLSchemaManager:  # pylint: disable=too-many-public-methods
    _extra_types: dict[str, GraphQLTypes] = {}
    _branch_details_by_name: dict[str, BranchDetails] = {}

    @classmethod
    def clear_cache(cls) -> None:
        cls._branch_details_by_name = {}

    @classmethod
    def _cache_branch(
        cls, branch: Branch, schema_branch: SchemaBranch, schema_hash: str | None = None
    ) -> BranchDetails:
        if not schema_hash:
            if branch.schema_hash:
                schema_hash = branch.schema_hash.main
            else:
                schema_hash = schema_branch.get_hash()
        branch_details = BranchDetails(
            branch_name=branch.name,
            schema_changed_at=Timestamp(branch.schema_changed_at) if branch.schema_changed_at else Timestamp(),
            schema_hash=schema_hash,
            gql_manager=cls(schema=schema_branch),
        )
        cls._branch_details_by_name[branch.name] = branch_details
        return branch_details

    @classmethod
    def get_manager_for_branch(cls, branch: Branch, schema_branch: SchemaBranch) -> GraphQLSchemaManager:
        if branch.name not in cls._branch_details_by_name:
            branch_details = cls._cache_branch(branch=branch, schema_branch=schema_branch)
            return branch_details.gql_manager
        cached_branch_details = cls._branch_details_by_name[branch.name]
        # try to use the schema_changed_at time b/c it is faster than checking the hash
        if branch.schema_changed_at:
            changed_at_time = Timestamp(branch.schema_changed_at)
            if changed_at_time > cached_branch_details.schema_changed_at:
                cached_branch_details = cls._cache_branch(branch=branch, schema_branch=schema_branch)
            return cached_branch_details.gql_manager
        if branch.schema_hash:
            current_hash = branch.active_schema_hash.main
        else:
            current_hash = schema_branch.get_hash()
        if cached_branch_details.schema_hash != current_hash:
            cached_branch_details = cls._cache_branch(
                branch=branch, schema_branch=schema_branch, schema_hash=current_hash
            )

        return cached_branch_details.gql_manager

    def __init__(self, schema: SchemaBranch) -> None:
        self.schema = schema

        self._full_graphql_schema: GraphQLSchema | None = None
        self._graphql_types: dict[str, GraphQLTypes] = {}

        self._load_attribute_types()
        if config.SETTINGS.experimental_features.graphql_enums:
            self._load_all_enum_types(node_schemas=self.schema.get_all().values())
        self._load_node_interface()

    def get_graphql_types(self) -> dict[str, GraphQLTypes]:
        return self._graphql_types

    def get_graphql_schema(
        self,
        include_query: bool = True,
        include_mutation: bool = True,
        include_subscription: bool = True,
        include_types: bool = True,
    ) -> GraphQLSchema:
        if all((include_query, include_mutation, include_subscription, include_types)):
            if not self._full_graphql_schema:
                self._full_graphql_schema = self.generate()
            return self._full_graphql_schema
        return self.generate(
            include_query=include_query,
            include_mutation=include_mutation,
            include_subscription=include_subscription,
            include_types=include_types,
        )

    def generate(
        self,
        include_query: bool = True,
        include_mutation: bool = True,
        include_subscription: bool = True,
        include_types: bool = True,
    ) -> GraphQLSchema:
        with SCHEMA_GENERATE_GRAPHQL_METRICS.labels(self.schema.name).time():
            query = self.get_gql_query() if include_query else None

            if include_types:
                if not include_query:
                    self.generate_object_types()
                types_dict = self.get_all()
                types = list(types_dict.values())
            else:
                types = []

            mutation = self.get_gql_mutation() if include_mutation else None
            subscription = None
            if include_subscription:
                partial_graphene_schema = graphene.Schema(
                    query=query,
                    mutation=mutation,
                    types=types,
                    auto_camelcase=False,
                    directives=DIRECTIVES,
                )
                subscription = self.get_gql_subscription(partial_graphql_schema=partial_graphene_schema.graphql_schema)

            graphene_schema = graphene.Schema(
                query=query,
                mutation=mutation,
                subscription=subscription,
                types=types,
                auto_camelcase=False,
                directives=DIRECTIVES,
            )

            return graphene_schema.graphql_schema

    def get_gql_query(self) -> type[InfrahubBaseQuery]:
        QueryMixin = self.generate_query_mixin()

        class Query(InfrahubBaseQuery, QueryMixin):  # type: ignore
            pass

        return Query

    def get_gql_mutation(self) -> type[InfrahubBaseMutation]:
        MutationMixin = self.generate_mutation_mixin()

        class Mutation(InfrahubBaseMutation, MutationMixin):  # type: ignore
            pass

        return Mutation

    def get_gql_subscription(self, partial_graphql_schema: graphene.Schema) -> type[InfrahubBaseSubscription]:
        class Subscription(InfrahubBaseSubscription):
            graphql_schema = partial_graphql_schema

        return Subscription

    def get_type(self, name: str) -> type[InfrahubObject]:
        if name in self._graphql_types and issubclass(
            self._graphql_types[name], (BaseAttributeType, graphene.Interface, graphene.ObjectType)
        ):
            return self._graphql_types[name]
        raise ValueError(f"Unable to find {name!r}")

    def get_mutation(self, name: str) -> type[InfrahubMutation]:
        if name in self._graphql_types and issubclass(self._graphql_types[name], InfrahubMutation):
            return self._graphql_types[name]
        raise ValueError(f"Unable to find {name!r}")

    def get_all(self) -> dict[str, GraphQLTypes]:
        infrahub_types = self._graphql_types
        infrahub_types.update(self._extra_types)
        return infrahub_types

    def set_type(self, name: str, graphql_type: GraphQLTypes) -> None:
        self._graphql_types[name] = graphql_type

    def _load_attribute_types(self) -> None:
        for data_type in ATTRIBUTE_TYPES.values():
            self.set_type(name=data_type.get_graphql_type_name(), graphql_type=data_type.get_graphql_type())

    def _load_node_interface(self) -> None:
        node_interface_schema = GenericSchema(
            name="Node", namespace="Core", description="Interface for all nodes in Infrahub"
        )
        interface = self.generate_interface_object(schema=node_interface_schema, populate_cache=True)
        edged_interface = self.generate_graphql_edged_object(
            schema=node_interface_schema, node=interface, populate_cache=True
        )
        self.generate_graphql_paginated_object(schema=node_interface_schema, edge=edged_interface, populate_cache=True)

    def _load_all_enum_types(self, node_schemas: Iterable[MainSchemaTypes]) -> None:
        for node_schema in node_schemas:
            self._load_enum_type(node_schema=node_schema)

    def _load_enum_type(self, node_schema: MainSchemaTypes) -> None:
        for attr_schema in node_schema.attributes:
            if not attr_schema.enum:
                continue
            base_enum_name = get_enum_attribute_type_name(node_schema, attr_schema)
            enum_value_name = f"{base_enum_name}Value"
            graphene_enum = generate_graphql_enum(name=enum_value_name, options=attr_schema.enum)
            data_type_class_name = f"{base_enum_name}EnumType"

            default_value = None
            if attr_schema.default_value:
                for g_enum in graphene_enum:
                    if g_enum.value == attr_schema.default_value:
                        default_value = g_enum.name
                        break

            graphene_field = graphene.Field(graphene_enum, default_value=default_value)
            create_class = type(f"{base_enum_name}AttributeCreate", (BaseAttributeCreate,), {"value": graphene_field})
            update_class = type(f"{base_enum_name}AttributeUpdate", (BaseAttributeUpdate,), {"value": graphene_field})
            data_type_class: type[InfrahubDataType] = type(
                data_type_class_name,
                (InfrahubDataType,),
                {
                    "label": data_type_class_name,
                    "graphql": graphene.String,
                    "graphql_query": TextAttributeType,
                    "graphql_create": create_class,
                    "graphql_update": update_class,
                    "graphql_filter": graphene_enum,
                    "infrahub": String,
                },
            )
            self.set_type(
                name=data_type_class.get_graphql_type_name(),
                graphql_type=data_type_class.get_graphql_type(),
            )
            ATTRIBUTE_TYPES[base_enum_name] = data_type_class

    def _get_related_input_type(self, relationship: RelationshipSchema) -> type[RelatedNodeInput]:
        peer_schema = self.schema.get(name=relationship.peer, duplicate=False)
        if (isinstance(peer_schema, NodeSchema) and peer_schema.is_ip_prefix()) or (
            isinstance(peer_schema, GenericSchema) and relationship.peer == InfrahubKind.IPPREFIX
        ):
            return RelatedPrefixNodeInput

        if (isinstance(peer_schema, NodeSchema) and peer_schema.is_ip_address()) or (
            isinstance(peer_schema, GenericSchema) and relationship.peer == InfrahubKind.IPADDRESS
        ):
            return RelatedIPAddressNodeInput

        return RelatedNodeInput

    def generate_object_types(self) -> None:  # pylint: disable=too-many-branches,too-many-statements
        """Generate all GraphQL objects for the schema and store them in the internal registry."""

        full_schema = self.schema.get_all(duplicate=False)

        # Generate all GraphQL Interface  Object first and store them in the registry
        for node_name, node_schema in full_schema.items():
            if not isinstance(node_schema, GenericSchema):
                continue
            interface = self.generate_interface_object(schema=node_schema, populate_cache=True)
            edged_interface = self.generate_graphql_edged_object(
                schema=node_schema, node=interface, populate_cache=True
            )
            self.generate_graphql_paginated_object(schema=node_schema, edge=edged_interface, populate_cache=True)

        # Define LineageSource and LineageOwner
        data_source = self.get_type(name=InfrahubKind.LINEAGESOURCE)
        data_owner = self.get_type(name=InfrahubKind.LINEAGEOWNER)
        self.define_relationship_property(data_source=data_source, data_owner=data_owner)
        relationship_property = self.get_type(name="RelationshipProperty")
        for data_type in ATTRIBUTE_TYPES.values():
            gql_type = self.get_type(name=data_type.get_graphql_type_name())
            gql_type._meta.fields["source"] = graphene.Field(data_source)
            gql_type._meta.fields["owner"] = graphene.Field(data_owner)

        # Generate all Nested, Edged and NestedEdged Interfaces and store them in the registry
        for node_name, node_schema in full_schema.items():
            if not isinstance(node_schema, GenericSchema):
                continue
            node_interface = self.get_type(name=node_name)

            nested_edged_interface = self.generate_nested_interface_object(
                schema=node_schema,
                base_interface=node_interface,
                relation_property=relationship_property,
            )

            nested_interface = self.generate_paginated_interface_object(
                schema=node_schema,
                base_interface=nested_edged_interface,
            )

            self.set_type(name=nested_interface._meta.name, graphql_type=nested_interface)
            self.set_type(name=nested_edged_interface._meta.name, graphql_type=nested_edged_interface)

        # Generate all GraphQL ObjectType, Nested, Paginated & NestedPaginated and store them in the registry
        for node_name, node_schema in full_schema.items():
            if isinstance(node_schema, (NodeSchema, ProfileSchema)):
                node_type = self.generate_graphql_object(schema=node_schema, populate_cache=True)
                node_type_edged = self.generate_graphql_edged_object(
                    schema=node_schema, node=node_type, populate_cache=True
                )
                nested_node_type_edged = self.generate_graphql_edged_object(
                    schema=node_schema, node=node_type, relation_property=relationship_property, populate_cache=True
                )

                self.generate_graphql_paginated_object(schema=node_schema, edge=node_type_edged, populate_cache=True)
                self.generate_graphql_paginated_object(
                    schema=node_schema, edge=nested_node_type_edged, nested=True, populate_cache=True
                )

        # Extend all types and related types with Relationships
        for node_name, node_schema in full_schema.items():
            node_type = self.get_type(name=node_name)

            for rel in node_schema.relationships:
                # Exclude hierarchical relationships, we will add them later
                if (
                    (isinstance(node_schema, NodeSchema) and node_schema.hierarchy)
                    or (isinstance(node_schema, GenericSchema) and node_schema.hierarchical)
                ) and rel.name in ("parent", "children", "ancestors", "descendants"):
                    continue

                peer_schema = self.schema.get(name=rel.peer, duplicate=False)
                if peer_schema.namespace == "Internal":
                    continue
                peer_filters = self.generate_filters(schema=peer_schema, top_level=False)

                if rel.cardinality == "one":
                    peer_type = self.get_type(name=f"NestedEdged{peer_schema.kind}")
                    node_type._meta.fields[rel.name] = graphene.Field(peer_type, resolver=single_relationship_resolver)

                elif rel.cardinality == "many":
                    peer_type = self.get_type(name=f"NestedPaginated{peer_schema.kind}")

                    if (isinstance(node_schema, NodeSchema) and node_schema.hierarchy) or (
                        isinstance(node_schema, GenericSchema) and node_schema.hierarchical
                    ):
                        peer_filters["include_descendants"] = graphene.Boolean()

                    node_type._meta.fields[rel.name] = graphene.Field(
                        peer_type, required=False, resolver=many_relationship_resolver, **peer_filters
                    )

            if (isinstance(node_schema, NodeSchema) and node_schema.hierarchy) or (
                isinstance(node_schema, GenericSchema) and node_schema.hierarchical
            ):
                if isinstance(node_schema, NodeSchema):
                    schema = self.schema.get(name=node_schema.hierarchy, duplicate=False)  # type: ignore[arg-type]
                    hierarchy_name = node_schema.hierarchy
                else:
                    schema = node_schema
                    hierarchy_name = node_schema.kind

                peer_filters = self.generate_filters(schema=schema, top_level=False)
                peer_type = self.get_type(name=f"NestedPaginated{hierarchy_name}")
                peer_type_edge = self.get_type(name=f"NestedEdged{hierarchy_name}")

                node_type._meta.fields["parent"] = graphene.Field(
                    peer_type_edge, required=False, resolver=single_relationship_resolver
                )
                node_type._meta.fields["children"] = graphene.Field(
                    peer_type, required=False, resolver=many_relationship_resolver, **peer_filters
                )
                node_type._meta.fields["ancestors"] = graphene.Field(
                    peer_type, required=False, resolver=ancestors_resolver, **peer_filters
                )
                node_type._meta.fields["descendants"] = graphene.Field(
                    peer_type, required=False, resolver=descendants_resolver, **peer_filters
                )

    def generate_query_mixin(self) -> type[object]:
        class_attrs = {}

        full_schema = self.schema.get_all(duplicate=False)

        # Generate all Graphql objectType and store internally
        self.generate_object_types()

        for node_name, node_schema in full_schema.items():
            if node_schema.namespace == "Internal":
                continue

            node_type = self.get_type(name=f"Paginated{node_name}")
            node_filters = self.generate_filters(schema=node_schema, top_level=True)

            class_attrs[node_schema.kind] = graphene.Field(
                node_type,
                resolver=default_paginated_list_resolver,
                **node_filters,
            )
            if node_name == InfrahubKind.GENERICACCOUNT:
                node_type = self.get_type(name=InfrahubKind.GENERICACCOUNT)
                class_attrs["AccountProfile"] = graphene.Field(node_type, resolver=account_resolver)

        return type("QueryMixin", (object,), class_attrs)

    def generate_mutation_mixin(self) -> type[object]:
        class_attrs: dict[str, Any] = {}

        full_schema = self.schema.get_all(duplicate=False)

        for node_schema in full_schema.values():
            if node_schema.namespace == "Internal":
                continue

            mutation_map: dict[str, type[InfrahubMutation]] = {
                InfrahubKind.ARTIFACTDEFINITION: InfrahubArtifactDefinitionMutation,
                InfrahubKind.REPOSITORY: InfrahubRepositoryMutation,
                InfrahubKind.READONLYREPOSITORY: InfrahubRepositoryMutation,
                InfrahubKind.PROPOSEDCHANGE: InfrahubProposedChangeMutation,
                InfrahubKind.GRAPHQLQUERY: InfrahubGraphQLQueryMutation,
                InfrahubKind.NAMESPACE: InfrahubIPNamespaceMutation,
                InfrahubKind.NUMBERPOOL: InfrahubNumberPoolMutation,
                InfrahubKind.MENUITEM: InfrahubCoreMenuMutation,
            }

            if isinstance(node_schema, NodeSchema) and node_schema.is_ip_prefix():
                base_class = InfrahubIPPrefixMutation
            elif isinstance(node_schema, NodeSchema) and node_schema.is_ip_address():
                base_class = InfrahubIPAddressMutation
            else:
                base_class = mutation_map.get(node_schema.kind, InfrahubMutation)

            if isinstance(node_schema, (NodeSchema, ProfileSchema)):
                mutations = self.generate_graphql_mutations(schema=node_schema, base_class=base_class)

                class_attrs[f"{node_schema.kind}Create"] = mutations.create.Field()
                class_attrs[f"{node_schema.kind}Update"] = mutations.update.Field()
                class_attrs[f"{node_schema.kind}Upsert"] = mutations.upsert.Field()
                class_attrs[f"{node_schema.kind}Delete"] = mutations.delete.Field()

            elif (
                isinstance(node_schema, GenericSchema)
                and (len(node_schema.attributes) + len(node_schema.relationships)) > 0
            ):
                graphql_mutation_update_input = self.generate_graphql_mutation_update_input(node_schema)
                update = self.generate_graphql_mutation_update(
                    schema=node_schema, base_class=base_class, input_type=graphql_mutation_update_input
                )
                self.set_type(name=update._meta.name, graphql_type=update)
                class_attrs[f"{node_schema.kind}Update"] = update.Field()

        return type("MutationMixin", (object,), class_attrs)

    def generate_graphql_object(self, schema: MainSchemaTypes, populate_cache: bool = False) -> type[InfrahubObject]:
        """Generate a GraphQL object Type from a Infrahub NodeSchema."""

        interfaces: set[type[InfrahubObject]] = set()

        if isinstance(schema, (NodeSchema, ProfileSchema)) and schema.inherit_from:
            for generic_name in schema.inherit_from:
                generic = self.get_type(name=generic_name)
                interfaces.add(generic)

        if isinstance(schema, NodeSchema):
            if not schema.inherit_from or InfrahubKind.GENERICGROUP not in schema.inherit_from:
                node_interface = self.get_type(name=InfrahubKind.NODE)
                interfaces.add(node_interface)

        meta_attrs = {
            "schema": schema,
            "name": schema.kind,
            "description": schema.description,
            "interfaces": interfaces,
        }

        main_attrs = {
            "id": graphene.Field(graphene.String, required=True, description="Unique identifier"),
            "hfid": graphene.Field(
                graphene.List(of_type=graphene.String), required=False, description="Human friendly identifier"
            ),
            "_updated_at": graphene.DateTime(required=False),
            "display_label": graphene.String(required=False),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        for attr in schema.local_attributes:
            attr_kind = get_attr_kind(schema, attr)
            attr_type = self.get_type(name=get_attribute_type(kind=attr_kind).get_graphql_type_name())
            main_attrs[attr.name] = graphene.Field(attr_type, required=not attr.optional, description=attr.description)

        graphql_object = type(schema.kind, (InfrahubObject,), main_attrs)

        if populate_cache:
            self.set_type(name=schema.kind, graphql_type=graphql_object)

        return graphql_object

    def generate_interface_object(
        self, schema: GenericSchema, populate_cache: bool = False
    ) -> type[graphene.Interface]:
        meta_attrs = {
            "name": schema.kind,
            "description": schema.description,
        }

        main_attrs = {
            "id": graphene.Field(graphene.String, required=False, description="Unique identifier"),
            "hfid": graphene.Field(
                graphene.List(of_type=graphene.String), required=False, description="Human friendly identifier"
            ),
            "display_label": graphene.String(required=False),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        for attr in schema.attributes:
            attr_kind = get_attr_kind(node_schema=schema, attr_schema=attr)
            attr_type = self.get_type(name=get_attribute_type(kind=attr_kind).get_graphql_type_name())
            main_attrs[attr.name] = graphene.Field(attr_type, required=not attr.optional, description=attr.description)

        interface_object = type(schema.kind, (InfrahubInterface,), main_attrs)

        if populate_cache:
            self.set_type(name=schema.kind, graphql_type=interface_object)

        return interface_object

    def define_relationship_property(self, data_source: type[InfrahubObject], data_owner: type[InfrahubObject]) -> None:
        type_name = "RelationshipProperty"

        meta_attrs = {
            "name": type_name,
            "description": "Defines properties for relationships",
        }

        main_attrs = {
            "is_visible": graphene.Boolean(required=False),
            "is_protected": graphene.Boolean(required=False),
            "updated_at": graphene.DateTime(required=False),
            "source": graphene.Field(data_source),
            "owner": graphene.Field(data_owner),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        relationship_property = type(type_name, (graphene.ObjectType,), main_attrs)

        self.set_type(name=type_name, graphql_type=relationship_property)

    def generate_graphql_mutations(
        self, schema: Union[NodeSchema, ProfileSchema], base_class: type[InfrahubMutation]
    ) -> GraphqlMutations:
        graphql_mutation_create_input = self.generate_graphql_mutation_create_input(schema)
        graphql_mutation_update_input = self.generate_graphql_mutation_update_input(schema)
        graphql_mutation_upsert_input = self.generate_graphql_mutation_upsert_input(schema)

        create = self.generate_graphql_mutation_create(
            schema=schema, base_class=base_class, input_type=graphql_mutation_create_input
        )
        upsert = self.generate_graphql_mutation_create(
            schema=schema,
            base_class=base_class,
            input_type=graphql_mutation_upsert_input,
            mutation_type="Upsert",
        )
        update = self.generate_graphql_mutation_update(
            schema=schema, base_class=base_class, input_type=graphql_mutation_update_input
        )
        delete = self.generate_graphql_mutation_delete(schema=schema, base_class=base_class)

        self.set_type(name=create._meta.name, graphql_type=create)
        self.set_type(name=update._meta.name, graphql_type=update)
        self.set_type(name=upsert._meta.name, graphql_type=upsert)
        self.set_type(name=delete._meta.name, graphql_type=delete)

        return GraphqlMutations(create=create, update=update, upsert=upsert, delete=delete)

    def generate_graphql_mutation_create_input(
        self, schema: Union[NodeSchema, ProfileSchema]
    ) -> type[graphene.InputObjectType]:
        """Generate an InputObjectType Object from a Infrahub NodeSchema

        Example of Object Generated by this function:
            class StatusCreateInput(InputObjectType):
                id = String(required=False)
                label = InputField(StringAttributeCreate, required=True)
                slug = InputField(StringAttributeCreate, required=True)
                description = InputField(StringAttributeCreate, required=False)
        """
        attrs: dict[str, Union[graphene.String, graphene.InputField]] = {"id": graphene.String(required=False)}

        for attr in schema.attributes:
            if attr.read_only:
                continue

            attr_kind = get_attr_kind(schema, attr)
            attr_type = get_attribute_type(kind=attr_kind).get_graphql_create()

            # A Field is not required if explicitly indicated or if a default value has been provided
            required = not attr.optional if not attr.default_value else False

            attrs[attr.name] = graphene.InputField(attr_type, required=required, description=attr.description)

        for rel in schema.relationships:
            if rel.internal_peer or rel.read_only:
                continue

            input_type = self._get_related_input_type(relationship=rel)

            required = not rel.optional
            if rel.cardinality == "one":
                attrs[rel.name] = graphene.InputField(input_type, required=required, description=rel.description)

            elif rel.cardinality == "many":
                attrs[rel.name] = graphene.InputField(
                    graphene.List(input_type), required=required, description=rel.description
                )

        return type(f"{schema.kind}CreateInput", (graphene.InputObjectType,), attrs)

    def generate_graphql_mutation_update_input(self, schema: MainSchemaTypes) -> type[graphene.InputObjectType]:
        """Generate an InputObjectType Object from a Infrahub NodeSchema

        Example of Object Generated by this function:
            class StatusUpdateInput(InputObjectType):
                id = String(required=False)
                hfid = InputField(List(of_type=String), required=False)
                label = InputField(StringAttributeUpdate, required=False)
                slug = InputField(StringAttributeUpdate, required=False)
                description = InputField(StringAttributeUpdate, required=False)
        """
        attrs: dict[str, Union[graphene.String, graphene.InputField]] = {
            "id": graphene.String(required=False),
            "hfid": graphene.List(of_type=graphene.String, required=False),
        }

        for attr in schema.attributes:
            if attr.read_only:
                continue
            attr_kind = get_attr_kind(schema, attr)
            attr_type = get_attribute_type(kind=attr_kind).get_graphql_update()
            attrs[attr.name] = graphene.InputField(attr_type, required=False, description=attr.description)

        for rel in schema.relationships:
            if rel.internal_peer or rel.read_only:
                continue

            input_type = self._get_related_input_type(relationship=rel)

            if rel.cardinality == "one":
                attrs[rel.name] = graphene.InputField(input_type, required=False, description=rel.description)

            elif rel.cardinality == "many":
                attrs[rel.name] = graphene.InputField(
                    graphene.List(input_type), required=False, description=rel.description
                )

        return type(f"{schema.kind}UpdateInput", (graphene.InputObjectType,), attrs)

    def generate_graphql_mutation_upsert_input(
        self, schema: Union[NodeSchema, ProfileSchema]
    ) -> type[graphene.InputObjectType]:
        """Generate an InputObjectType Object from a Infrahub NodeSchema

        Example of Object Generated by this function:
            class StatusUpsertInput(InputObjectType):
                id = String(required=False)
                hfid = InputField(List(of_type=String), required=False)
                label = InputField(StringAttributeUpdate, required=True)
                slug = InputField(StringAttributeUpdate, required=True)
                description = InputField(StringAttributeUpdate, required=False)
        """
        attrs: dict[str, Union[graphene.String, graphene.InputField]] = {
            "id": graphene.String(required=False),
            "hfid": graphene.List(of_type=graphene.String, required=False),
        }

        for attr in schema.attributes:
            if attr.read_only:
                continue

            attr_kind = get_attr_kind(schema, attr)
            attr_type = get_attribute_type(kind=attr_kind).get_graphql_update()

            # A Field is not required if explicitly indicated or if a default value has been provided
            required = not attr.optional if not attr.default_value else False

            attrs[attr.name] = graphene.InputField(attr_type, required=required, description=attr.description)

        for rel in schema.relationships:
            if rel.internal_peer or rel.read_only:
                continue

            input_type = self._get_related_input_type(relationship=rel)

            required = not rel.optional
            if rel.cardinality == "one":
                attrs[rel.name] = graphene.InputField(input_type, required=required, description=rel.description)

            elif rel.cardinality == "many":
                attrs[rel.name] = graphene.InputField(
                    graphene.List(input_type), required=required, description=rel.description
                )

        return type(f"{schema.kind}UpsertInput", (graphene.InputObjectType,), attrs)

    def generate_graphql_mutation_create(
        self,
        schema: Union[NodeSchema, ProfileSchema],
        input_type: type[graphene.InputObjectType],
        base_class: type[InfrahubMutation] = InfrahubMutation,
        mutation_type: str = "Create",
    ) -> type[InfrahubMutation]:
        """Generate a GraphQL Mutation to CREATE an object based on the specified NodeSchema."""
        name = f"{schema.kind}{mutation_type}"

        object_type = self.generate_graphql_object(schema=schema)

        main_attrs: dict[str, Any] = {"ok": graphene.Boolean(), "object": graphene.Field(object_type)}

        meta_attrs: dict[str, Any] = {"schema": schema, "name": name, "description": schema.description}
        main_attrs["Meta"] = type("Meta", (object,), meta_attrs)

        args_attrs = {
            "data": input_type(required=True),
        }
        main_attrs["Arguments"] = type("Arguments", (object,), args_attrs)

        return type(name, (base_class,), main_attrs)

    def generate_graphql_mutation_update(
        self,
        schema: MainSchemaTypes,
        input_type: type[graphene.InputObjectType],
        base_class: type[InfrahubMutation] = InfrahubMutation,
    ) -> type[InfrahubMutation]:
        """Generate a GraphQL Mutation to UPDATE an object based on the specified NodeSchema."""
        name = f"{schema.kind}Update"

        object_type = self.generate_graphql_object(schema=schema)

        main_attrs: dict[str, Any] = {"ok": graphene.Boolean(), "object": graphene.Field(object_type)}

        meta_attrs: dict[str, Any] = {"schema": schema, "name": name, "description": schema.description}
        main_attrs["Meta"] = type("Meta", (object,), meta_attrs)

        args_attrs = {
            "data": input_type(required=True),
        }
        main_attrs["Arguments"] = type("Arguments", (object,), args_attrs)

        return type(name, (base_class,), main_attrs)

    @staticmethod
    def generate_graphql_mutation_delete(
        schema: Union[NodeSchema, ProfileSchema], base_class: type[InfrahubMutation] = InfrahubMutation
    ) -> type[InfrahubMutation]:
        """Generate a GraphQL Mutation to DELETE an object based on the specified NodeSchema."""
        name = f"{schema.kind}Delete"

        main_attrs: dict[str, Any] = {"ok": graphene.Boolean()}

        meta_attrs = {"schema": schema, "name": name, "description": schema.description}
        main_attrs["Meta"] = type("Meta", (object,), meta_attrs)

        args_attrs: dict[str, Any] = {
            "data": DeleteInput(required=True),
        }
        main_attrs["Arguments"] = type("Arguments", (object,), args_attrs)

        return type(name, (base_class,), main_attrs)

    def generate_filters(
        self, schema: MainSchemaTypes, top_level: bool = False, include_properties: bool = True
    ) -> dict[str, Union[graphene.Scalar, graphene.List]]:
        """Generate the GraphQL filters for a given Schema object.

        The generated filter will be different if we are at the top_level (query)
        or if we are generating the filter for a relationship inside a node.

        At the top, level it will be possible to query with a list of ID
        Inside a node, it's only possible to query with a single ID

        Args:
            schema (Union[NodeSchema, GenericSchema]): Schema to generate the filters
            top_level (bool, optional): Flag to indicate if are at the top level or not. Defaults to False.

        Returns:
            dict: A Dictionary containing all the filters with their name as the key and their Type as value
        """

        filters: dict[str, Any] = {"offset": graphene.Int(), "limit": graphene.Int(), "order": OrderInput()}
        default_filters: list[str] = list(filters.keys())

        filters["ids"] = graphene.List(graphene.ID)
        if not top_level:
            filters["isnull"] = graphene.Boolean()

        if schema.human_friendly_id and top_level:
            # HFID filter limited to top level because we can't filter on HFID for relationships (yet)
            filters["hfid"] = graphene.List(graphene.String)

        for attr in schema.attributes:
            attr_kind = get_attr_kind(node_schema=schema, attr_schema=attr)
            filters.update(
                get_attribute_type(kind=attr_kind).get_graphql_filters(
                    name=attr.name, include_properties=include_properties, include_isnull=top_level
                )
            )

        if top_level:
            filters.update(get_attribute_type().get_graphql_filters(name="any"))
            filters["partial_match"] = graphene.Boolean()

        if not top_level:
            return filters

        for rel in schema.relationships:
            peer_schema = self.schema.get(name=rel.peer, duplicate=False)

            if peer_schema.namespace == "Internal":
                continue

            if rel.kind == RelationshipKind.GROUP:
                peer_filters = self.generate_filters(schema=peer_schema, top_level=False, include_properties=False)
            else:
                peer_filters = self.generate_filters(schema=peer_schema, top_level=False)

            for key, value in peer_filters.items():
                if key in default_filters:
                    continue
                filters[f"{rel.name}__{key}"] = value

        return filters

    def generate_graphql_edged_object(
        self,
        schema: MainSchemaTypes,
        node: type[InfrahubObject],
        relation_property: Optional[type[InfrahubObject]] = None,
        populate_cache: bool = False,
    ) -> type[InfrahubObject]:
        """Generate a edged GraphQL object Type from a Infrahub NodeSchema for pagination."""

        object_name = f"Edged{schema.kind}"
        if relation_property:
            object_name = f"NestedEdged{schema.kind}"

        meta_attrs: dict[str, Any] = {
            "schema": schema,
            "name": object_name,
            "description": schema.description,
            "interfaces": set(),
        }

        main_attrs: dict[str, Any] = {
            "node": graphene.Field(node, required=False),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        if relation_property:
            main_attrs["properties"] = graphene.Field(relation_property)

        graphql_edged_object = type(object_name, (InfrahubObject,), main_attrs)

        if populate_cache:
            self.set_type(name=object_name, graphql_type=graphql_edged_object)

        return graphql_edged_object

    def generate_graphql_paginated_object(
        self, schema: MainSchemaTypes, edge: type[InfrahubObject], nested: bool = False, populate_cache: bool = False
    ) -> type[InfrahubObject]:
        """Generate a paginated GraphQL object Type from a Infrahub NodeSchema."""

        object_name = f"Paginated{schema.kind}"
        if nested:
            object_name = f"NestedPaginated{schema.kind}"

        meta_attrs: dict[str, Any] = {
            "schema": schema,
            "name": object_name,
            "description": schema.description,
            "default_resolver": default_resolver,
            "interfaces": set(),
        }

        main_attrs: dict[str, Any] = {
            "count": graphene.Int(required=False),
            "edges": graphene.List(of_type=edge),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        if isinstance(schema, (NodeSchema, GenericSchema)):
            main_attrs["permissions"] = graphene.Field(
                PaginatedObjectPermission, required=True, resolver=parent_field_name_resolver
            )

        graphql_paginated_object = type(object_name, (InfrahubObject,), main_attrs)

        if populate_cache:
            self.set_type(name=object_name, graphql_type=graphql_paginated_object)

        return graphql_paginated_object

    def generate_nested_interface_object(
        self,
        schema: GenericSchema,
        relation_property: graphene.ObjectType,
        base_interface: graphene.ObjectType,
        populate_cache: bool = False,
    ) -> type[InfrahubObject]:
        meta_attrs: dict[str, Any] = {
            "name": f"NestedEdged{schema.kind}",
            "schema": schema,
            "description": schema.description,
        }

        main_attrs: dict[str, Any] = {
            "node": graphene.Field(base_interface, required=False),
            "_updated_at": graphene.DateTime(required=False),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        if relation_property:
            main_attrs["properties"] = graphene.Field(relation_property)

        object_name = f"NestedEdged{schema.kind}"
        nested_interface_object = type(object_name, (InfrahubObject,), main_attrs)

        if populate_cache:
            self.set_type(name=object_name, graphql_type=nested_interface_object)

        return nested_interface_object

    @staticmethod
    def generate_paginated_interface_object(
        schema: GenericSchema, base_interface: type[graphene.ObjectType]
    ) -> type[InfrahubObject]:
        meta_attrs: dict[str, Any] = {
            "name": f"NestedPaginated{schema.kind}",
            "schema": schema,
            "description": schema.description,
        }

        main_attrs: dict[str, Any] = {
            "count": graphene.Int(required=False),
            "edges": graphene.List(of_type=base_interface),
            "Meta": type("Meta", (object,), meta_attrs),
        }

        return type(f"NestedPaginated{schema.kind}", (InfrahubObject,), main_attrs)



from __future__ import annotations

import asyncio
import logging
import traceback
from collections.abc import Coroutine
from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, NoReturn, TypeVar

import pendulum
import typer
from click.exceptions import Exit
from httpx import HTTPError
from pendulum.datetime import DateTime
from rich.console import Console
from rich.logging import RichHandler
from rich.markup import escape

from ..ctl.exceptions import FileNotValidError, QueryNotFoundError
from ..exceptions import (
    AuthenticationError,
    Error,
    GraphQLError,
    NodeNotFoundError,
    ResourceNotDefinedError,
    SchemaNotFoundError,
    ServerNotReachableError,
    ServerNotResponsiveError,
)
from ..yaml import YamlFile
from .client import initialize_client_sync

if TYPE_CHECKING:
    from ..schema.repository import InfrahubRepositoryConfig

YamlFileVar = TypeVar("YamlFileVar", bound=YamlFile)
T = TypeVar("T")


def init_logging(debug: bool = False) -> None:
    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)
    logging.getLogger("httpx").setLevel(logging.ERROR)
    logging.getLogger("httpcore").setLevel(logging.ERROR)

    log_level = "DEBUG" if debug else "INFO"
    FORMAT = "%(message)s"
    logging.basicConfig(level=log_level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])
    logging.getLogger("infrahubctl")


def handle_exception(exc: Exception, console: Console, exit_code: int) -> NoReturn:
    """Handle exeception in a different fashion based on its type."""
    if isinstance(exc, Exit):
        raise typer.Exit(code=exc.exit_code)
    if isinstance(exc, AuthenticationError):
        console.print(f"[red]Authentication failure: {exc!s}")
        raise typer.Exit(code=exit_code)
    if isinstance(exc, (ServerNotReachableError, ServerNotResponsiveError)):
        console.print(f"[red]{exc!s}")
        raise typer.Exit(code=exit_code)
    if isinstance(exc, HTTPError):
        console.print(f"[red]HTTP communication failure: {exc!s} on {exc.request.method} to {exc.request.url}")
        raise typer.Exit(code=exit_code)
    if isinstance(exc, GraphQLError):
        print_graphql_errors(console=console, errors=exc.errors)
        raise typer.Exit(code=exit_code)
    if isinstance(exc, (SchemaNotFoundError, NodeNotFoundError, ResourceNotDefinedError)):
        console.print(f"[red]Error: {exc!s}")
        raise typer.Exit(code=exit_code)

    console.print(f"[red]Error: {exc!s}")
    console.print(traceback.format_exc())
    raise typer.Exit(code=exit_code)


def catch_exception(
    console: Console | None = None, exit_code: int = 1
) -> Callable[[Callable[..., T]], Callable[..., T | Coroutine[Any, Any, T]]]:
    """Decorator to handle exception for commands."""
    if not console:
        console = Console()

    def decorator(func: Callable[..., T]) -> Callable[..., T | Coroutine[Any, Any, T]]:
        if asyncio.iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args: Any, **kwargs: Any) -> T:
                try:
                    return await func(*args, **kwargs)
                except (Error, Exception) as exc:  # pylint: disable=broad-exception-caught
                    return handle_exception(exc=exc, console=console, exit_code=exit_code)

            return async_wrapper

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> T:
            try:
                return func(*args, **kwargs)
            except (Error, Exception) as exc:  # pylint: disable=broad-exception-caught
                return handle_exception(exc=exc, console=console, exit_code=exit_code)

        return wrapper

    return decorator


def execute_graphql_query(
    query: str,
    variables_dict: dict[str, Any],
    repository_config: InfrahubRepositoryConfig,
    branch: str | None = None,
    debug: bool = False,
) -> dict:
    console = Console()
    query_object = repository_config.get_query(name=query)
    query_str = query_object.load_query()

    client = initialize_client_sync()
    response = client.execute_graphql(
        query=query_str,
        branch_name=branch,
        variables=variables_dict,
        raise_for_error=False,
    )

    if debug:
        console.print("-" * 40)
        console.print(f"Response for GraphQL Query {query}")
        console.print(response)
        console.print("-" * 40)

    return response


def print_graphql_errors(console: Console, errors: list) -> None:
    if not isinstance(errors, list):
        console.print(f"[red]{escape(str(errors))}")

    for error in errors:
        if isinstance(error, dict) and "message" in error and "path" in error:
            console.print(f"[red]{escape(str(error['path']))} {escape(str(error['message']))}")
        else:
            console.print(f"[red]{escape(str(error))}")


def parse_cli_vars(variables: list[str] | None) -> dict[str, str]:
    if not variables:
        return {}

    return {var.split("=")[0]: var.split("=")[1] for var in variables if "=" in var}


def calculate_time_diff(value: str) -> str | None:
    """Calculate the time in human format between a timedate in string format and now."""
    try:
        time_value = pendulum.parse(value)
    except pendulum.parsing.exceptions.ParserError:
        return None

    if not isinstance(time_value, DateTime):
        return None

    pendulum.set_locale("en")
    return time_value.diff_for_humans(other=pendulum.now(), absolute=True)


def find_graphql_query(name: str, directory: str | Path = ".") -> str:
    if isinstance(directory, str):
        directory = Path(directory)

    for query_file in directory.glob("**/*.gql"):
        if query_file.stem != name:
            continue
        return query_file.read_text(encoding="utf-8")

    raise QueryNotFoundError(name=name)


def render_action_rich(value: str) -> str:
    if value == "created":
        return f"[green]{value.upper()}[/green]"
    if value == "updated":
        return f"[magenta]{value.upper()}[/magenta]"
    if value == "deleted":
        return f"[red]{value.upper()}[/red]"

    return value.upper()


def get_fixtures_dir() -> Path:
    """Get the directory which stores fixtures that are common to multiple unit/integration tests."""
    here = Path(__file__).resolve().parent
    return here.parent.parent / "tests" / "fixtures"


def load_yamlfile_from_disk_and_exit(
    paths: list[Path], file_type: type[YamlFileVar], console: Console
) -> list[YamlFileVar]:
    has_error = False
    try:
        data_files = file_type.load_from_disk(paths=paths)
    except FileNotValidError as exc:
        console.print(f"[red]{exc.message}")
        raise typer.Exit(1) from exc

    for data_file in data_files:
        if data_file.valid and data_file.content:
            continue
        console.print(f"[red]{data_file.error_message} ({data_file.location})")
        has_error = True

    if has_error:
        raise typer.Exit(1)

    return data_files



from __future__ import annotations

import random
import socket
import time
from pathlib import Path
from uuid import UUID

from .utils import base16encode

BASE = 16
DIVISOR = BASE - 1
CHARACTERS = list("0123456789abcdefghijklmnopqrstuvwxyz")[:BASE]
HOSTNAME = socket.gethostname()
DEFAULT_NAMESPACE = str(Path(__file__).parent.resolve())

# Code inspired from https://github.com/isaacharrisholt/uuidt


def generate_uuid() -> str:
    return str(UUIDT())


def encode_number(number: int, min_length: int) -> str:
    """Encode a number into a base16 string and ensure the result has a minimum size.
    If the initial response produced doesn't match the min requirement,
    random number will be used to fill the gap
    """
    response = base16encode(number=number).lower()
    if len(response) >= min_length:
        return response
    return response + "".join(random.choices(CHARACTERS, k=min_length - len(response)))


class UUIDT:
    def __init__(
        self,
        namespace: str | None = None,
        timestamp: int | None = None,
        hostname: str | None = None,
        random_chars: str | None = None,
    ) -> None:
        self.namespace = namespace or DEFAULT_NAMESPACE
        self.timestamp = timestamp or time.time_ns()
        self.hostname = hostname or HOSTNAME
        self.random_chars = random_chars or "".join(random.choices(CHARACTERS, k=8))

    def __str__(self) -> str:
        hostname_enc = sum(self.hostname.encode("utf-8"))
        namespace_enc = sum(self.namespace.encode("utf-8"))

        timestamp_str = encode_number(number=self.timestamp, min_length=16)
        hostname_str = encode_number(number=hostname_enc, min_length=4)
        namespace_str = encode_number(number=namespace_enc, min_length=4)

        return f"{timestamp_str[:8]}-{timestamp_str[8:12]}-{timestamp_str[-4:]}-{hostname_str[:4]}-{namespace_str[:4]}{self.random_chars[:8]}"

    def short(self) -> str:
        """Return the last 8 digit of the UUID (the most random part)"""
        return str(self)[-8:]

    @classmethod
    def new(cls, namespace: str | None = None) -> UUID:
        return UUID(str(cls(namespace=namespace)))
